#include "loginscreen.h"

LoginScreen::LoginScreen(QWidget *parent ) : QWidget( parent )
{
	Shares::smbUmountAll();
	Shares::sshUmountAll();

	networkManager = new NetworkManager();
	wifiMenu = nullptr;
	ethernetMenu = nullptr;
	vpnMenu = nullptr;
	networkTimer = new QTimer();
	connect( networkTimer, SIGNAL( timeout() ), this, SLOT( updateNetwork() ) );
	networkTimer->start( 1000 );

	screen = qApp->primaryScreen();
	setGeometry(screen->geometry());
	QIcon icon(":/icons/other/librework.icns");
	setWindowIcon( icon );
	setWindowTitle(tr("Librework Desktop"));
	setWindowFlags( windowFlags() | Qt::FramelessWindowHint );
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	clientSession = 0;
#endif

	qsrand( QTime::currentTime().msec() );

	pressedKeysTimer = new QTimer();
	pressedKeysTimer->setSingleShot( true );
	connect( pressedKeysTimer, SIGNAL(timeout()), this, SLOT( clearPressedKeys() ) );

	initInitScreen();
	initUserControls();
	showUserControls( false );

	connect(passwordLineEdit, SIGNAL(returnPressed()), this, SLOT(loginButtonClicked()));
	connect(usernameLineEdit, SIGNAL(returnPressed()), this, SLOT(loginButtonClicked()));
	connect(loginButton, SIGNAL(clicked(bool)), this, SLOT(loginButtonClicked()));
	connect(exitButton, SIGNAL(customContextMenuRequested(QPoint)), exitButton, SLOT(showMenu()));
	connect(wifiButton, SIGNAL(clicked()), this, SLOT(showWifiMenu()));

	pixmapFrame->show();
	pixmapFrame->setFocusPolicy( Qt::StrongFocus );
	pixmapFrame->setFocus();

	loginScreenTimeout = new QTimer( this );
	int interval = Settings::getValue( "LoginScreenTimeout", LOGINSCREEN_DEFAULT_TIMEOUT).toInt();
	loginScreenTimeout->setInterval( 1000 * interval );

	connect( animation, SIGNAL( finished() ), this, SLOT( animationFinished() ) );
	connect( pixmapFrame, SIGNAL( mousePressed() ), this, SLOT( mouseOrKeyboardAction() ) );
	connect( pixmapFrame, SIGNAL( mouseMoved() ), this, SLOT( mouseOrKeyboardAction() ) );
	connect( pixmapFrame, SIGNAL( keyPressed(QKeyEvent*) ), this, SLOT( mouseOrKeyboardAction(QKeyEvent*) ) );
	connect( pixmapFrame, SIGNAL( escapePressed() ), this, SLOT( doInitScreen() ) );

	connect( loginScreenTimeout, SIGNAL( timeout() ), this, SLOT( doInitScreen() ) );
	connect( pixmapFrame, SIGNAL( mousePressed() ), loginScreenTimeout, SLOT( start() ) );
	connect( pixmapFrame, SIGNAL( mouseMoved() ), loginScreenTimeout, SLOT( start() ) );
	connect( pixmapFrame, SIGNAL( keyPressed(QKeyEvent*) ), loginScreenTimeout, SLOT( start() ) );
	connect( usernameLineEdit, SIGNAL( textChanged(const QString &) ), loginScreenTimeout, SLOT( start() ) );
	connect( passwordLineEdit, SIGNAL( textChanged(const QString &) ), loginScreenTimeout, SLOT( start() ) );
	connect( serverComboBox, SIGNAL( activated(int) ), loginScreenTimeout, SLOT( start() ) );
	connect( serverComboBox, SIGNAL( highlighted(int) ), loginScreenTimeout, SLOT( start() ) );
	connect( serverComboBox, SIGNAL( currentIndexChanged(int) ), loginScreenTimeout, SLOT( start() ) );

	QString autoLogin = Settings::getValue( "AutologinUser" ).toString();
	if ( ! autoLogin.isEmpty() )
	{
		QString autoLoginSession = Settings::getValue( "AutologinSession" ).toString();
	}

	QTimer::singleShot( 250, this, SLOT( setDateTime() ) );

	timer = new QTimer();
	timer->setInterval( 1000 );
	connect( timer, SIGNAL( timeout() ), this, SLOT( setDateTime() ) );
	timer->start();
}

LoginScreen::~LoginScreen()
{
	delete networkManager;
	networkTimer->stop();
	delete networkTimer;
	timer->stop();
	animation->stop();
	loginScreenTimeout->stop();
	delete animation;
	delete timer;
	delete pressedKeysTimer;
	delete loginScreenTimeout;
	delete logoLabel;
	delete usernameLineEdit;
	delete passwordLineEdit;
	delete serverComboBox;
	delete loginButton;
	delete wifiButton;
	delete ethernetButton;
	delete vpnButton;
	delete exitButton;
	delete pixmapFrame;
	if ( clientSession != nullptr )
	{
		delete clientSession;
	}
}

void LoginScreen::initInitScreen()
{
	QSize desktopSize = screen->geometry().size();
	int paddingLeft = desktopSize.width() / 60;
	int paddingRight = desktopSize.width() / 75;
	int menuHeight = desktopSize.height() / 30;
	int fontPixelSize = desktopSize.height() / 50;
	menuStyleSheet = QString(
		"QMenu"
		"{"
		"	background-color: rgba( 0, 0, 0, 0.5 );"
		"	border-color: rgba( 255, 255, 255, 0.5 );"
		"	border-width: 3px;"
		"	border-style: solid;"
		"	border-radius: 8px;"
		"	padding: 10px;"
		"	font: 75 Bold %2px '%3';"
		"	color: #ededed;"
		"}"
		"QMenu::item"
		"{"
		"	height: %1px;"
		"	margin: 0px;"
		"	padding-left: %4px;"
		"	padding-right: %5px;"
		"}"
		"QMenu::item:selected"
		"{"
		"	background-color: rgba( 127, 127, 127, 0.5 );"
		"}"
		"QMenu::item:hover"
		"{"
		"	background-color: rgba( 127, 127, 127, 0.5 );"
		"}"
		).arg( menuHeight ).arg( fontPixelSize ).arg( LIBREWORK_FONT ).arg(paddingLeft).arg(paddingRight);
	QString configurationFile = selectConfigurationFile();

	QPixmap pixmap = selectPixmap( configurationFile );

	pixmapFrame = new PixmapFrame( this, pixmap );
	pixmapFrame->setColor( QColor(0,0,0) );
	int pX = 5;
	int pY = 75;

	animation = new QPropertyAnimation(pixmapFrame, "opacity");
	animation->setDuration(600);
	animation->setStartValue( 1.0 );
	animation->setEndValue( LOGINSCREEN_OBSCURED_OPACITY );

	if ( ! configurationFile.isEmpty() )
	{
		QSettings settings( configurationFile, QSettings::IniFormat );
		pixmapFrame->setFontColor( QColor( settings.value( "LoginScreen/FontColor" , "#ffffff" ).toString() ) );
		pX = settings.value( "LoginScreen/TextPositionX", pX ).toInt();
		pY = settings.value( "LoginScreen/TextPositionY", pY ).toInt();
	}
	pixmapFrame->setTextPos( pX * screen->geometry().width() / 100, pY * screen->geometry().height() / 100 );

	pixmapFrame->setFixedSize( screen->geometry().size() );

	pixmapFrame->setFontFamily( LIBREWORK_FONT );
	pixmapFrame->setTimeFontPixelSize( screen->geometry().width() / 15);
	pixmapFrame->setDateFontPixelSize( screen->geometry().width() / 25);
	pixmapFrame->resize();

	QList<QScreen*> screens = qApp->screens();
	int counter = 0;
	for ( int i = 0; i < screens.count(); i++ )
	{
		qDebug("Screen number: %d, name: '%s' (primary name: '%s')", i, screens[i]->name().toUtf8().data(), screen->name().toUtf8().data());
		if ( screens[i]->name() != screen->name() )
		{
			qDebug("Showing pixmap on screen %s with geometry x: %d, y: %d, width: %d, height: %d", screens[i]->name().toUtf8().data(), screens[i]->geometry().x(), screens[i]->geometry().y(), screens[i]->geometry().width(), screens[i]->geometry().height());
			background.append( new QLabel( nullptr, Qt::FramelessWindowHint ) );
			background[counter]->setAttribute( Qt::WA_DeleteOnClose, true );
			background[counter]->setPixmap( pixmap );
			background[counter]->setScaledContents( true );
			background[counter]->setGeometry( screens[i]->geometry() );
			background[counter]->setFixedSize( screens[i]->geometry().size() );
			background[counter]->show();
			counter++;
		}
	}
}

void LoginScreen::initUserControls()
{
	QSize desktopSize = screen->geometry().size();

	QSize logoLabelSize = QSize( desktopSize.width() / 4, desktopSize.width() / 12 );
	QPoint logoLabelPos = QPoint( ( desktopSize.width() - logoLabelSize.width() ) / 2, desktopSize.height() * 30 / 100 );
	QString logoLabelStyleSheet(
		"QLabel"
		"{"
		"	background-color: rgba( 255, 255, 255, 0.45 );"
		"	border-width: 0px;"
		"	border-style: solid;"
		"	border-radius: 20px;"
		"}"
		);	
	logoLabel = new QLabel( pixmapFrame );
	logoLabel->setFocusPolicy( Qt::NoFocus );
	logoLabel->setStyleSheet( logoLabelStyleSheet );
	logoLabel->setScaledContents( false );
	logoLabel->setAlignment( Qt::AlignCenter );
	logoLabel->setFixedSize( logoLabelSize );
	QPixmap logoLabelPixmap = QPixmap( ":/icons/tuxumos.png" );
	logoLabel->setPixmap( logoLabelPixmap.scaledToHeight( logoLabelSize.height() * 2.8, Qt::SmoothTransformation ) );
	logoLabel->setGeometry( logoLabelPos.x(), logoLabelPos.y(), logoLabelSize.width(), logoLabelSize.height() );

	QSize usernameLineEditSize( desktopSize.width() / 4, desktopSize.height() / 15 );

	int spacing = desktopSize.height() / 50;
	QPoint usernameLineEditPos( ( desktopSize.width() - usernameLineEditSize.width() ) / 2, logoLabelPos.y() + logoLabelSize.height() + spacing );
	usernameLineEdit = new QLineEdit( pixmapFrame );
	usernameLineEdit->setMaxLength( 255 );
	usernameLineEdit->setPlaceholderText( tr("User") );
	usernameLineEdit->setAlignment(Qt::AlignCenter);
	usernameLineEdit->setFixedSize( usernameLineEditSize );
	usernameLineEdit->setGeometry( usernameLineEditPos.x(), usernameLineEditPos.y(), usernameLineEditSize.width(), usernameLineEditSize.height() );
	QString lastLogin = Settings::getValue("LastLogin").toString(); // 'General' section must not be specified
	usernameLineEdit->setText( lastLogin );
	usernameLineEditTextChanged( lastLogin );
	connect( usernameLineEdit, SIGNAL(textChanged(QString)), this, SLOT(usernameLineEditTextChanged(QString)));

	QSize passwordLineEditSize( desktopSize.width() / 4, desktopSize.height() / 15 );
	QPoint passwordLineEditPos( ( desktopSize.width() - passwordLineEditSize.width() ) / 2, usernameLineEditPos.y() + usernameLineEditSize.height() + spacing );
	passwordLineEdit = new QLineEdit( pixmapFrame );
	passwordLineEdit->setMaxLength( 255 );
	passwordLineEdit->setPlaceholderText( tr("Password") );
	passwordLineEdit->setAlignment(Qt::AlignCenter);
	QString passwordStyleSheet = QString(
		"QLineEdit"
		"{"
		"	background-color: rgba( 0, 0, 0, 0.5 );"
		"	border-color: rgba( 255, 255, 255, 0.5 );"
		"	border-width: 3px;"
		"	border-style: solid;"
		"	border-radius: 8px;"
		"	font: 75 Bold %1px '%2';"
		"	color: #ededed;"
		"}"
		).arg(desktopSize.width() / 64).arg( LIBREWORK_FONT );
	passwordLineEdit->setEchoMode(QLineEdit::Password);
	passwordLineEdit->setStyleSheet( passwordStyleSheet );
	passwordLineEdit->setFixedSize( passwordLineEditSize );
	passwordLineEdit->setGeometry( passwordLineEditPos.x(), passwordLineEditPos.y(), passwordLineEditSize.width(), passwordLineEditSize.height() );

	QSize serverComboBoxSize( desktopSize.width() / 5, desktopSize.height() / 25 );
	QPoint serverComboBoxPos( ( desktopSize.width() - serverComboBoxSize.width() ) / 2, passwordLineEditPos.y() + passwordLineEditSize.height() + spacing );
	serverComboBox = new QComboBox( pixmapFrame );

	qDebug("1");
	sessionSettings = SessionSetting::getSessionSettings();

	qDebug("2");
	QStringList macAddresses = getMacAddresses();
	QString lastSession = Settings::getValue("LastSession").toString();
	QString lastLoginSession;

	for ( int i = 0; i < sessionSettings.count(); i++ )
	{
		qDebug("i: %d", i);
		serverComboBox->addItem( sessionSettings[i].icon(), sessionSettings[i].name(), i );
		qDebug("1");
		for ( int j = 0; j < macAddresses.count(); j++ )
		{
			qDebug("j: %d", j);
			QStringList match = sessionSettings[i].computers().sessionKeys().filter( macAddresses[j], Qt::CaseInsensitive );
			if ( match.count() > 0 )
			{
				lastSession = sessionSettings[i].name();
			}

			match = sessionSettings[i].computers().userKeys().filter( macAddresses[j], Qt::CaseInsensitive );
			if ( match.count() > 0 )
			{
				lastLoginSession = sessionSettings[i].name();
			}
		}
	}
	qDebug("3");

	connect( serverComboBox, SIGNAL(currentTextChanged(QString)), this, SLOT(serverChanged(QString)));

	qDebug("LastSession: '%s'", lastSession.toUtf8().data());
	if ( serverComboBox->findText( lastSession, Qt::MatchExactly ) != -1 )
	{
		serverComboBox->setCurrentText( lastSession );
	}

	if ( ! lastLoginSession.isEmpty() )
	{
		serverComboBox->setCurrentText( lastLoginSession );
	}

	QFont arial( LIBREWORK_FONT );
	arial.setPixelSize( desktopSize.width() / 60 );
	serverComboBox->setFont( arial );
	QString serverStyleSheet = QString(
		"QComboBox"
		"{"
		"	background-color: rgba( 0, 0, 0, 0.5 );"
		"	border-color: rgba( 255, 255, 255, 0 );"
		"	border-width: 0px;"
		"	border-style: solid;"
		"	border: none;"
		"	padding-left: 10px;"
		"	padding-right: 10px;"
		"	selection-background-color: rgba( 127, 127, 127, 0.5);"
		"	font: 75 Bold %1px '%2';"
		"	font-size: %1px;"
		"	icon-size: %1px;"
		"	border-radius: 8px;"
		"	color: #ededed;"
		"}"
		).arg(desktopSize.width() / 72).arg( LIBREWORK_FONT );
	serverComboBox->setStyleSheet( serverStyleSheet );
	serverComboBox->setFixedSize( serverComboBoxSize );
	serverComboBox->setGeometry( serverComboBoxPos.x(), serverComboBoxPos.y(), serverComboBoxSize.width(), serverComboBoxSize.height() );

	QString loginScreenButtonStyle = QString(
		"QPushButton"
		"{"
		"	background-color: transparent;"
		"	border: none;"
		"	icon-size: %1px;"
		"}"
		"QPushButton:selected"
		"{"
		"	background-color: transparent;"
		"	border: none;"
		"	icon-size: %1px;"
		"}"
		"QPushButton:pressed:hover"
		"{"
		"	background-color: transparent;"
		"	border: none;"
		"       padding-top: 5px;"
		"       padding-left: 5px;"
		"	icon-size: %1px;"
		"}"
		"QPushButton:hover"
		"{"
		"	background-color: rgba( 255, 255, 255, 25 );"
		"	border-width: 0px;"
		"	border-style: solid;"
		"	border-radius: %2px;"
		"	icon-size: %1px;"
		"};"
		).arg(desktopSize.width() / 40).arg(desktopSize.height() / 10);
	QSize buttonSize( desktopSize.width() / 25, desktopSize.width() / 25 );

	QPoint loginButtonPos( passwordLineEditPos.x() + passwordLineEditSize.width() + 5, passwordLineEditPos.y() );
	loginButton = new QPushButton( QIcon( ":/icons/logon.png" ), "", pixmapFrame );
	loginButton->setFocusPolicy( Qt::NoFocus );
	loginButton->setFlat( true );
	loginButton->setStyleSheet( loginScreenButtonStyle );
	loginButton->setGeometry( loginButtonPos.x(), loginButtonPos.y(), buttonSize.width(), buttonSize.height() );

	QPoint exitButtonPos( desktopSize.width() - buttonSize.width() - spacing, desktopSize.height() - buttonSize.height() - spacing );
	exitButton = new QPushButton( QIcon( ":/icons/logoff.png" ), "", pixmapFrame );
	exitButton->setFocusPolicy( Qt::NoFocus );
	exitButton->setFlat( true );
	exitButton->setStyleSheet( loginScreenButtonStyle );
	exitButton->setGeometry( exitButtonPos.x(), exitButtonPos.y(), buttonSize.width(), buttonSize.height() );

	QPoint wifiButtonPos( desktopSize.width() - 2 * buttonSize.width() - 2 * spacing, desktopSize.height() - buttonSize.height() - spacing );
	wifiButton = new QPushButton( QIcon( ":/icons/wifi.png" ), "", pixmapFrame );
	wifiButton->setFocusPolicy( Qt::NoFocus );
	wifiButton->setFlat( true );
	wifiButton->setStyleSheet( loginScreenButtonStyle );
	wifiButton->setGeometry( wifiButtonPos.x(), wifiButtonPos.y(), buttonSize.width(), buttonSize.height() );

	QPoint ethernetButtonPos( desktopSize.width() - 4 * buttonSize.width() - 4 * spacing, desktopSize.height() - buttonSize.height() - spacing );
	ethernetButton = new QPushButton( QIcon( ":/icons/lan.png" ), "", pixmapFrame );
	ethernetButton->setFocusPolicy( Qt::NoFocus );
	ethernetButton->setFlat( true );
	ethernetButton->setStyleSheet( loginScreenButtonStyle );
	ethernetButton->setGeometry( ethernetButtonPos.x(), ethernetButtonPos.y(), buttonSize.width(), buttonSize.height() );

	setShutdownMenu();
}

void LoginScreen::serverChanged(QString server)
{
	ClientSettings * clientSettings = new ClientSettings();
	QString _userName = clientSettings->getLogin( server );
	if ( ! _userName.isEmpty() )
	{
		usernameLineEdit->setText( clientSettings->getLogin( server ) );
	} else {
		usernameLineEdit->selectAll();
	}
	delete clientSettings;
}

void LoginScreen::usernameLineEditTextChanged(QString text)
{
	QSize desktopSize = screen->geometry().size();
	int borderWidth = 0;
	if ( text.isEmpty() )
	{
		borderWidth = 3;
	}

	QString usernameStyleSheet = QString(
		"QLineEdit"
		"{"
		"	background-color: transparent;"
		"	border-color: rgba( 255, 255, 255, 0.5 );"
		"	border-width: %1px;"
		"	border-style: solid;"
		"	border-radius: 8px;"
		"	font: 75 Bold %2px '%3';"
		"	color: #ededed;"
		"}"
		"QLineEdit:focus"
		"{"
		"	background-color: rgba( 0, 0, 0, 0.5 );"
		"	border-color: rgba( 255, 255, 255, 0.5 );"
		"	border-width: 3px;"
		"	border-style: solid;"
		"	border-radius: 8px;"
		"	font: 75 Bold %2px '%3';"
		"	color: #ededed;"
		"}"
		).arg(borderWidth).arg(desktopSize.width() / 48).arg( LIBREWORK_FONT );
	usernameLineEdit->setStyleSheet( usernameStyleSheet );
}

void LoginScreen::showUserControls(bool show)
{
	if ( show )
	{
		if ( ! Settings::getValue( "SmartCardLogin", false ).toBool() )
		{
			logoLabel->show();
			usernameLineEdit->show();
			serverComboBox->show();
		}
		passwordLineEdit->show();
		if ( ! passwordLineEdit->text().isEmpty() )
		{
			loginButton->show();
		}
		exitButton->show();
		wifiButton->show();
		ethernetButton->show();
		connect( passwordLineEdit, SIGNAL(textChanged(const QString &)), this, SLOT(passwordTextChanged(const QString &) ));
	} else {
		logoLabel->hide();
		usernameLineEdit->hide();
		passwordLineEdit->hide();
		serverComboBox->hide();
		loginButton->hide();
		exitButton->hide();
		wifiButton->hide();
		ethernetButton->hide();
		disconnect( passwordLineEdit, SIGNAL(textChanged(const QString &)), this, SLOT(passwordTextChanged(const QString &) ));
	}
}

void LoginScreen::passwordTextChanged(const QString & )
{
	if ( ! passwordLineEdit->text().isEmpty() )
	{
		loginButton->show();
	} else {
		loginButton->hide();
	}
}

void LoginScreen::mouseOrKeyboardAction(QKeyEvent * event)
{
	if ( ! Settings::getValue( "SmartCardLogin", false ).toBool() )
	{
		if ( event != nullptr )
		{
			pressedKeys = event->text();
		}
		doAnimation();
	} else {
		if ( event != nullptr )
		{
			if ( event->key() == Qt::Key_Return || event->key() == Qt::Key_Enter )
			{
				pressedKeysTimer->stop();
				if ( checkPressedKeys() )
				{
					doSmartCardLogin( pressedKeys );
				} else {
					clearPressedKeys();
				}
			} else {
				if ( ! pressedKeysTimer->isActive() )
				{
					pressedKeysTimer->start( PRESSEDKEYSTIMER );
				}
				pressedKeys += event->text();
			}
		}
	}
}

void LoginScreen::clearPressedKeys()
{
	pressedKeys.clear();
}

bool LoginScreen::checkPressedKeys()
{
	if ( pressedKeys.size() < 8 )
	{
		return false;
	}

	bool ret = true;
	for ( int i = 0; i < pressedKeys.size(); i++ )
	{
		if ( ! pressedKeys.at( i ).isDigit() )
		{
			ret = false;
		}
	}
	return ret;
}

void LoginScreen::doAnimation()
{
	pixmapFrame->setFocusPolicy( Qt::NoFocus );
	passwordLineEdit->clear();
	disconnect( pixmapFrame, SIGNAL( mousePressed() ), this, SLOT( mouseOrKeyboardAction() ) );
	disconnect( pixmapFrame, SIGNAL( mouseMoved() ), this, SLOT( mouseOrKeyboardAction() ) );
	disconnect( pixmapFrame, SIGNAL( keyPressed(QKeyEvent*) ), this, SLOT( mouseOrKeyboardAction(QKeyEvent*) ) );
	timer->stop();

	showUserControls( true );

	if ( usernameLineEdit->text().isEmpty() )
	{
		usernameLineEdit->setFocus();
	} else {
		passwordLineEdit->setFocus();
	}

	if ( ! pressedKeys.isEmpty() )
	{
		if ( usernameLineEdit->text().isEmpty() )
		{
			usernameLineEdit->setText( pressedKeys );
		} else {
			passwordLineEdit->setText( pressedKeys );
		}
	}

	animation->start();
}

void LoginScreen::doInitScreen()
{
	passwordLineEdit->clear();
	clearPressedKeys();
	pixmapFrame->setFocusPolicy( Qt::StrongFocus );
	pixmapFrame->setFocus();
	showUserControls( false );
	animation->stop();
	pixmapFrame->setOpacity( 1.0 );
	pixmapFrame->showText( true );
	timer->start();
	connect( pixmapFrame, SIGNAL( mousePressed() ), this, SLOT( mouseOrKeyboardAction() ) );
	connect( pixmapFrame, SIGNAL( mouseMoved() ), this, SLOT( mouseOrKeyboardAction() ) );
	connect( pixmapFrame, SIGNAL( keyPressed(QKeyEvent*) ), this, SLOT( mouseOrKeyboardAction(QKeyEvent*) ) );
}

void LoginScreen::animationFinished()
{
	pixmapFrame->showText( false );
	loginScreenTimeout->start();
}

QString LoginScreen::selectConfigurationFile()
{
	QString ret;
	QDir wallpaperDir( SHAREDIR + QString(QDir::separator()) + "/loginscreen" );
	QStringList wallpaperConfigList = wallpaperDir.entryList( QStringList( "*.ini" ) );
	if ( wallpaperConfigList.count() > 0 )
	{
		int index = qrand() % wallpaperConfigList.count();
		ret = wallpaperDir.absolutePath() + QDir::separator() + wallpaperConfigList[index];
	}
	return ret;
}

void LoginScreen::setDateTime()
{
	QString time = QTime::currentTime().toString( "hh:mm" );
	QString date = QDate::currentDate().toString( "dddd, d MMM" );
	date.replace( 0, 1, date[0].toUpper() );
	pixmapFrame->setText( time, date );
}

QPixmap LoginScreen::selectPixmap( QString configurationFile)
{
	QPixmap pixmap;
	if ( ! configurationFile.isEmpty() )
	{
		QSettings settings( configurationFile, QSettings::IniFormat );
		QRect screenSize = screen->geometry();
		QString sizeString = QString::number( screenSize.width() ) + "x" + QString::number( screenSize.height() );
		if ( settings.value( sizeString + "/File", "").toString().isEmpty() ) // In case not exact resolution found
		{
			// Find best aspect ratio file
			double aspectRatio = (double) screenSize.width() / (double) screenSize.height();
			double bestAspectRatioDiff = -1;
			QStringList childGroups = settings.childGroups();
			for ( int i = 0; i < childGroups.count(); i++ )
			{
				bool ok;
				if ( childGroups[i] != "LoginScreen" )
				{
					int width = childGroups[i].section("x", 0, 0).toInt(&ok);
					if ( ! ok )
					{
						break;
					}
					int height = childGroups[i].section("x", 1, 1).toInt(&ok);
					if ( ok && width >= screenSize.width() && height >= screenSize.height() )
					{
						double pixmapAspectRatio = (double) width / (double) height;
						double diff = pixmapAspectRatio - aspectRatio;
						if ( diff < 0 )
						{
							diff = -diff;
						}
						if ( diff < bestAspectRatioDiff || bestAspectRatioDiff == -1 )
						{
							bestAspectRatioDiff = diff;
							sizeString = childGroups[i];
						}
					}
				}
			}
		}
		QString pixmapFile = settings.value( sizeString + "/File", settings.value( "LoginScreen/File", "" ) ).toString();
		if ( ! pixmapFile.isEmpty() )
		{
			pixmap = QPixmap( SHAREDIR + QString( QDir::separator() ) + "loginscreen" + QString( QDir::separator() ) + pixmapFile );
		}
	} else {
		QDir wallpaperDir( SHAREDIR + QString(QDir::separator()) + "/loginscreen" );
		QStringList extensions;
		extensions << "*.png";
		extensions << "*.jpg";
		extensions << "*.gif";
		QStringList wallpaperList = wallpaperDir.entryList( extensions );
		if ( wallpaperList.count() > 0 )
		{
			int index = qrand() % wallpaperList.count();
			pixmap = QPixmap( wallpaperDir.absolutePath() + QDir::separator() + wallpaperList[index] );
		}
	}
	return pixmap;
}

void LoginScreen::setShutdownMenu()
{
	QSize desktopSize = screen->geometry().size();
	QMenu * menu = new QMenu;
	menu->setStyle( new MenuStyle(desktopSize.width() / 100) );
	menu->addAction(QIcon(":/icons/logoff.png"), tr("Power Off"), Power::instance(), SLOT(powerOff()));
	menu->addAction(QIcon(":/icons/reboot.png"), tr("Reboot"), Power::instance(), SLOT(reboot()));
	menu->addAction(QIcon(":/icons/suspend.png"), tr("Suspend"), Power::instance(), SLOT(suspend()));
	menu->addAction(QIcon(":/icons/hibernate.png"), tr("Hibernate"), Power::instance(), SLOT(hibernate()));

	menu->setStyleSheet( menuStyleSheet );
	exitButton->setMenu( menu );
}

void LoginScreen::showWifiMenu()
{
	QApplication::setOverrideCursor(Qt::WaitCursor);

	QSize desktopSize = screen->geometry().size();
	if ( wifiMenu != nullptr )
	{
		delete wifiMenu;
	}
	wifiMenu = new QMenu;
	wifiMenu->setStyle( new MenuStyle(desktopSize.width() / 100) );
	QList<WifiAp> networks = networkManager->apList();
	QString selectedNetwork = networkManager->connectedAp();

	qDebug("Networks: %d", networks.count());

	for ( int i = 0; i < networks.count(); i++ )
	{
		QAction * action;
		QString text = networks[i].essid + " (" + QString::number( networks[i].quality ) + " %)";
		qDebug("WifiAp: %s", text.toUtf8().data());
		if ( networks[i].essid == selectedNetwork )
		{
			action = wifiMenu->addAction(QIcon(":/icons/wifi-green.png"), text);
		} else {
			action = wifiMenu->addAction(QIcon(":/icons/wifi.png"), text);
		}
		action->setData( networks[i].essid );
	}

	connect( wifiMenu, SIGNAL(triggered(QAction*)), this, SLOT(selectWifi(QAction*)) );

	QApplication::setOverrideCursor(Qt::ArrowCursor);

	wifiMenu->setStyleSheet( menuStyleSheet );
	QPoint position( desktopSize.width(),  desktopSize.height() );
	wifiMenu->exec( position );
}

void LoginScreen::updateNetwork()
{
	if ( networkManager->wifiInterfaces().count() > 0 )
	{
		wifiButton->show();
		if ( networkManager->isWifiConnected() )
		{
			wifiButton->setIcon( QIcon( ":/icons/wifi-on.png" ) );
		} else {
			wifiButton->setIcon( QIcon( ":/icons/wifi-off.png" ) );
		}
	} else {
		wifiButton->hide();
	}
}

void LoginScreen::selectWifi(QAction * action)
{
	QList<QNetworkInterface> interfaces = networkManager->wifiInterfaces();
	if ( interfaces.count() == 0 )
	{
		return;
	}
	bool ok;
	QString key = QInputDialog::getText( nullptr, tr("Wireless connection"), tr("Please, introduce key for network \"%1\"").arg(action->data().toString()), QLineEdit::Normal, QString(), &ok, Qt::Popup, Qt::ImhNone);
	qApp->processEvents();
	if ( ok )
	{
		QApplication::setOverrideCursor(Qt::WaitCursor);
		int ret = networkManager->connectToAp( action->data().toString(), key );
		QApplication::setOverrideCursor(Qt::ArrowCursor);
		if ( ret != SUCCESS_CODE )
		{
			if ( ret == DHCP_ERROR )
			{
				QMessageBox::critical(nullptr, tr("Wireless connection"), tr("Wireless connection failed. No DHCP offers available"));
			}
			if ( ret == WPA_SUPPLICANT_ERROR )
			{
				QMessageBox::critical(nullptr, tr("Wireless connection"), tr("Wireless connection failed. Authentication or WPA error"));
			}
		}
	}
}

void LoginScreen::doSmartCardLogin(QString cardNumber)
{
	setCursor( Qt::WaitCursor );

	QString smartUser;
	int smartLibreworkServerPort = 0;
	QStringList error;

	QString smartSession = Settings::getValue( "SmartCardSession", "" ).toString();
	if ( smartSession.isEmpty() )
	{
		error.append( tr("Undefined parameter") );
		error.append( tr("The parameter 'smartCardSession' is not defined on librework-displaymanager.ini" ) );
	}
		
	QString smartLibreworkServer = Settings::getValue( "SmartCardLibreworkServer", "" ).toString();
	if ( smartSession.isEmpty() || smartLibreworkServer.isEmpty() )
	{
		if ( error.count() == 0 )
		{
			error.append( tr("Undefined parameter") );
			error.append( tr("The parameter 'smartCardLibreworkServer' is not defined on librework-displaymanager.ini" ) );
		}
	} else {
		smartLibreworkServerPort = smartLibreworkServer.section( ':', 1, 1 ).toInt();
		if ( smartLibreworkServerPort == 0 )
		{
			smartLibreworkServerPort = LIBREWORK_SERVER_DEFAULT_PORT;
		}
		smartLibreworkServer = smartLibreworkServer.section( ':', 0, 0 );
		qDebug("Connecting to server %s:%d", smartLibreworkServer.toUtf8().data(), smartLibreworkServerPort);
		if ( SessionManager::connectToServer( smartLibreworkServer, smartLibreworkServerPort ) )
		{
			qDebug("Connected. Getting user name for id %s", cardNumber.toUtf8().data());
			smartUser = LWClient::sendMessage( "get_user_from_card?" + cardNumber );
			if ( smartUser.isEmpty() )
			{
				qDebug("Error getting user name from id");
				error.append( tr("Invalid card") );
				error.append( tr("There is no user assigned to this smart card") );
			} else {
				qDebug("Got '%s' user name", smartUser.toUtf8().data());
			}
		} else {
			error.append( tr("Connection error") );
			error.append( tr("Cannot connect to librework server on %1:%2").arg(smartLibreworkServer).arg(smartLibreworkServerPort) );
		}
	}

	setCursor( Qt::ArrowCursor );

	if ( error.count() == 2 )
	{
		QMessageBox::critical( 0, error[0], error[1] );
		disconnect( pixmapFrame, SIGNAL( mousePressed() ), this, SLOT( mouseOrKeyboardAction() ) );
		disconnect( pixmapFrame, SIGNAL( mouseMoved() ), this, SLOT( mouseOrKeyboardAction() ) );
		disconnect( pixmapFrame, SIGNAL( keyPressed(QKeyEvent*) ), this, SLOT( mouseOrKeyboardAction(QKeyEvent*) ) );
		doInitScreen();
	} else {
		qDebug("No error. Getting password...");
		usernameLineEdit->setText( smartUser );
		serverComboBox->setCurrentText( smartSession );
		showUserControls( true );
		passwordLineEdit->setFocus();
	}
}

void LoginScreen::loginButtonClicked()
{
	if ( usernameLineEdit->text().isEmpty() || passwordLineEdit->text().isEmpty() )
	{
		return;
	}

	setCursor( Qt::WaitCursor );
	pixmapFrame->setOpacity( LOGINSCREEN_OBSCURED_OPACITY );
	qApp->processEvents();

	showUserControls( false );
	qApp->processEvents();

	QPropertyAnimation * fadeOut = new QPropertyAnimation( this ); // LoginScreen is its parent, so it takes care of its deletion
	fadeOut->setTargetObject( pixmapFrame );
	fadeOut->setPropertyName( "opacity" );
	fadeOut->setDuration( FADE_OUT_TIME );
	fadeOut->setStartValue( LOGINSCREEN_OBSCURED_OPACITY );
	fadeOut->setEndValue( 0.0 );

	for ( int i = 0; i < background.count(); i++ )
	{
		background[i]->close();
		delete background[i];
	}

	qDebug("Setting last server: '%s'", serverComboBox->currentText().toUtf8().data());

	if ( Settings::getValue("RememberSession").toBool() )
	{
		Settings::setValue( "LastSession", serverComboBox->currentText()); // 'General' section must not be specified
	}

	QStringList error;

	username = usernameLineEdit->text();
	QString serverConfig = serverComboBox->currentText();

	int index = serverComboBox->currentData().toInt();
	connect( fadeOut, SIGNAL(finished()), this, SLOT(hideSlot()));

	error = SessionManager::launch( usernameLineEdit->text(), passwordLineEdit->text(), sessionSettings[index], this );

	if ( fadeOut->currentValue().toDouble() == 0.0 )
	{
		hide();
	}

	if ( error.count() == 2 )
	{
		disconnect( fadeOut, SIGNAL(finished()), this, SLOT(hideSlot()));
		QMessageBox::critical( 0, error[0], error[1] );
		clearPressedKeys();
		doInitScreen();
	} else {
		fadeOut->start();
		qApp->processEvents();
		passwordLineEdit->clear();
		clearPressedKeys();
	}
	setCursor( Qt::ArrowCursor );
	qApp->processEvents();
}

QStringList LoginScreen::getMacAddresses()
{
	QStringList macAddresses;
	foreach(QNetworkInterface netInterface, QNetworkInterface::allInterfaces())
	{
		if ( ! (netInterface.flags() & QNetworkInterface::IsLoopBack) )
		{
			macAddresses += netInterface.hardwareAddress();
		}
	}
	return macAddresses;
}

void LoginScreen::forceTranslations()
{
	QString ok = tr("OK");
	Q_UNUSED(ok);
	QString cancel = tr("Cancel");
	Q_UNUSED(cancel);
}

void LoginScreen::finishSession( int status )
{
	Q_UNUSED( status )

	Shares::smbUmountAll();
	Shares::sshUmountAll();
	SessionManager::umountHome();

        close();
        deleteLater();
}

void LoginScreen::hideSlot()
{
	this->hide();
}
