#include "pam.h"

#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
static pam_handle_t *user_auth_handle = NULL; // this gets set by pam_start
static pam_handle_t *greeter_auth_handle = NULL; // this gets set by pam_start
struct pam_response * reply = nullptr;
#endif

#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
int Pam::functionConversation(int numMsg, const struct pam_message **msg, struct pam_response ** resp, void *appdata_ptr)
{
	Q_UNUSED(appdata_ptr);
	Q_UNUSED(numMsg);
	Q_UNUSED(msg);
	for (int i = 0; i < numMsg; i++)
	{
		qDebug("msg: %s", msg[i]->msg);
	}
	*resp = reply;

	return PAM_SUCCESS;
}
#endif

bool Pam::startPam( bool greeter, QString service, QString username )
{
	Logger::logDebug(QString("Starting PAM with service %1 for username %2").arg(service).arg(username));
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	reply = (struct pam_response *)malloc(sizeof(struct pam_response));

	const struct pam_conv local_conversation = { functionConversation, NULL };

	pam_handle_t * handle;


	int retval;

	retval = pam_start( service.toUtf8().data(), username.toUtf8().data(), &local_conversation, &handle);

	if ( greeter )
	{
		greeter_auth_handle = handle;
	} else {
		user_auth_handle = handle;
	}

	if ( retval == PAM_SUCCESS )
	{
		QString xdgSessionClass = "user";
		if ( greeter )
		{
			xdgSessionClass = "greeter";
		}
		int val = pam_misc_setenv(handle, "XDG_SESSION_CLASS", xdgSessionClass.toUtf8().data(), 0);
		Logger::logInfo(QString("XDG_SESSION_CLASS=%1 (PAM_SUCCESS=%2): %3").arg(xdgSessionClass).arg( PAM_SUCCESS ).arg( retval ) );

		return true;
	} else {
		Logger::logError("Cannot start PAM.");
		stopPam( greeter, 1 );
		return false;
	}
#endif
}

bool Pam::stopPam(bool greeter, int retval)
{
	pam_handle_t * handle;
	if ( greeter )
	{
		handle = greeter_auth_handle;
	} else {
		handle = user_auth_handle;
	}
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	if ( reply == nullptr && handle == NULL )
	{
		return true;
	}
	int ret = pam_end( handle, retval );
	reply = nullptr;
	if ( greeter )
	{
		greeter_auth_handle = NULL;
	} else {
		user_auth_handle = NULL;
	}
	return ret;
#endif
	return true;
}

bool Pam::authenticateSystemUser( QString username, QString password )
{
	Logger::logDebug(QString("Authenticating user %1").arg(username));
#ifdef Q_OS_WIN
	//LogonUser(L"LocalServer", L"NT AUTHORITY", NULL, LOGON32_LOGON_SERVICE, LOGON32_PROVIDER_DEFAULT, &hToken)
	return true;
#endif
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	Logger::logDebug("Checking greeter_auth_handle...");
	if ( greeter_auth_handle == NULL )
	{
		Logger::logDebug("Is null. Returning false");
		return false;
	}
	Logger::logDebug("greeter_auth_handle ok");

	reply[0].resp = strdup( password.toUtf8().data() );
	reply[0].resp_retcode = 0;
	Logger::logDebug("Doing pam_authenticate");
	int retval = pam_authenticate(greeter_auth_handle, 0);
	Logger::logDebug("Doing pam_set_item");
	pam_set_item(greeter_auth_handle, PAM_RUSER, username.toUtf8().data());

	if (retval != PAM_SUCCESS)
	{
		Logger::logDebug("pam_set_item error. Cannot authenticate.");
		return false;
	}

	Logger::logDebug("pam_set_item success. Authentication successful.");
	return true;
#endif
}

bool Pam::openSession( QString xdgSessionClass, QString seat )
{
	pam_handle_t * handle;
	if ( xdgSessionClass == "greeter" )
	{
		handle = greeter_auth_handle;
	} else {
		handle = user_auth_handle;
	}
	bool ok = true;
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	int retval = pam_open_session(handle, 0 /*PAM_SILENT*/);
	if ( retval == PAM_SUCCESS )
	{
		Logger::logInfo(QString("pam_open_session: PAM_SUCCESS"));
	} else {
		ok = false;
		Logger::logInfo(QString("pam_open_session error: (PAM_SUCCESS=%1): %2").arg( PAM_SUCCESS ).arg( retval ) );
	}

	retval = pam_misc_setenv(handle, "XDG_SEAT", seat.toUtf8().data(), 0);
	if ( retval == PAM_SUCCESS )
	{
		Logger::logInfo(QString("XDG_SEAT=%1 (PAM_SUCCESS)").arg( seat ) );
	} else {
		ok = false;
		Logger::logInfo(QString("error setting XDG_SEAT=%1 (PAM_SUCCESS=%2): %3").arg( seat ).arg( PAM_SUCCESS ).arg( retval ) );
	}

#if defined(Q_OS_LINUX)
	QFile tty( "/sys/class/tty/tty0/active" );
	if ( tty.open( QIODevice::ReadOnly ) )
	{
		Logger::logDebug("Successfuly opened tty file");
		QByteArray ttyByteArray = tty.readAll();
		ttyByteArray.replace( "tty", "" );
		bool convOk;
		int ttyInt = ttyByteArray.toInt( &convOk );
		if ( convOk )
		{
			retval = pam_misc_setenv(handle, "XDG_VTNR", QString::number( ttyInt ).toUtf8().data(), 0);
			if ( retval == PAM_SUCCESS )
			{
				Logger::logInfo(QString("XDG_VTNR=%1 (PAM_SUCCESS)").arg( ttyInt ) );
			} else {
				ok = false;
				Logger::logInfo(QString("Error setting XDG_VTNR=%1 (PAM_SUCCESS=%2): %3").arg( ttyInt ).arg( PAM_SUCCESS ).arg( retval ) );
			}
		}
	} else {
		Logger::logWarning("Cannot open active tty file");
		ok = false;
	}
#endif
#endif
	return ok;
}

bool Pam::closeSession(bool greeter)
{
	pam_handle_t * handle;
	if ( greeter )
	{
		handle = greeter_auth_handle;
	} else {
		handle = user_auth_handle;
	}
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	int retval = pam_close_session(handle, 0 /*PAM_SILENT*/);
	return ( stopPam( greeter, retval ) == PAM_SUCCESS );
#endif
	return true;
}
