#include "session.h"

#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
static pam_handle_t *local_auth_handle = NULL; // this gets set by pam_start
struct pam_response * reply;
#endif

void Session::getUserAndGroupIds( QString username, int * uid, int * gid )
{
#ifdef Q_OS_WIN
	Q_UNUSED( username );
	Q_UNUSED( uid );
	Q_UNUSED( gid );
#endif
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	struct passwd pwd;
	struct passwd *result;
	char *buf;
	size_t bufsize;
	int s;

	bufsize = sysconf(_SC_GETPW_R_SIZE_MAX);
	if (bufsize == (size_t)-1)          /* Value was indeterminate */
	{
		bufsize = 16384;        /* Should be more than enough */
	}

	buf = (char*)malloc(bufsize);
	if (buf == NULL)
	{
		* uid = -1;
		* gid = -1;
		qDebug("Error getting memory for buffer");
		return;
	}

	s = getpwnam_r(username.toUtf8().data(), &pwd, buf, bufsize, &result);
	if (result == NULL)
	{
		if (s == 0)
		{
			qDebug("Not found\n");
		} else {
			qDebug("Error %d on getpwnam_r", s);
		}
		* uid = -1;
		* gid = -1;
	} else {
		* uid = result->pw_uid;
		* gid = result->pw_gid;
	}
	free( buf );
#endif
} 

int Session::setCurrentDir(const char *dirname)
{
	QDir::setCurrent(dirname);
#ifdef Q_OS_WIN
	if (SetCurrentDirectoryA(dirname))
	{
		return 0;
	} else {
		return -1;
	}
#endif

#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	return chdir(dirname);
#endif
}

bool Session::createSystemUser( int uid, QString homes, QString username, QString password )
{
#ifdef Q_OS_WIN
	Q_UNUSED( uid );
	Q_UNUSED( homes );
	Q_UNUSED( username );
	Q_UNUSED( password );
#endif
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	int _uid = uid;
	bool ret = false;
	bool exists = false;
	int lastUserId = 1000;
	QFile passwd(LIBREWORK_PASSWD_FILE);
	if (passwd.open(QIODevice::ReadOnly))
	{
		QTextStream in(&passwd);
		while (!in.atEnd())
		{
			QString line = in.readLine();
			QString lineUsername = line.section(":", 0, 0);
			int lineUserId = line.section(":", 2, 2).toInt();
			if ( lineUserId > lastUserId )
			{
				lastUserId = lineUserId;
			}
			if ( lineUserId == _uid )
			{
				exists = true;
				if ( lineUsername == username )
				{
					ret = true;
				}
			}
		}
		passwd.close();
	}

	if ( _uid == -1 )
	{
		_uid = lastUserId + 1;
	}

	if ( ! exists )
	{
		QFile passwd(LIBREWORK_PASSWD_FILE);
		passwd.open(QIODevice::WriteOnly | QIODevice::Append);
		QByteArray line;
		line += username;
		line += ":x:"; // No password entry
		line += QString::number( _uid );
		line += ":50::"; // Staff group
		line += homes + "/" + username;
		line += ":/bin/bash\n";
		if ( passwd.write( line ) == -1 )
		{
			Logger::logError( QString( "Cannot write to file %1" ).arg( LIBREWORK_PASSWD_FILE ) );
			ret = false;
		} else {
			ret = true;
		}
		passwd.close();

		QStringList libreworkUserGroups = Settings::getValue("Advanced/LibreworkNSSUserGroups").toStringList();
		QFile lwGroup(LIBREWORK_GROUP_FILE);
		if ( ret && lwGroup.open(QIODevice::WriteOnly | QIODevice::Append) )
		{
			QFile group("/etc/group");
			if ( group.open(QIODevice::ReadOnly))
			{
				QTextStream in(&group);
				while ( ret && ! in.atEnd() )
				{
					QString line = in.readLine();
					QString lineGroupName = line.section(":", 0, 0);
					QString lineGroupId = line.section(":", 2, 2);
					if ( libreworkUserGroups.contains( lineGroupName ) )
					{
						QByteArray groupLine = QString( lineGroupName + ":x:" + lineGroupId + ":" + username + "\n" ).toUtf8();
						if ( lwGroup.write( groupLine ) == -1 )
						{
							Logger::logError( QString( "Cannot write to file %1" ).arg( LIBREWORK_GROUP_FILE ) );
							ret = false;
						}
					}
				}
				group.close();
			}
			lwGroup.close();
		}
		QFile lwShadow(LIBREWORK_SHADOW_FILE);
		if ( ret && lwShadow.open(QIODevice::WriteOnly | QIODevice::Append) )
		{
			lwShadow.setPermissions( QFileDevice::ReadOwner );
			QByteArray line = username.toUtf8() + ":" + QCryptographicHash::hash( password.toUtf8(), QCryptographicHash::Sha512 ) + ":19999:0:99999:7:::\n";
			if ( lwShadow.write( line ) == -1 )
			{
				Logger::logError( QString( "Cannot write to file %1" ).arg( LIBREWORK_SHADOW_FILE ) );
				ret = false;
			} else {
				ret = true;
			}
			lwShadow.close();
		}
	}

	return ret;
#endif
	return false;
}

#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
int Session::functionConversation(int numMsg, const struct pam_message **msg, struct pam_response ** resp, void *appdata_ptr)
{
	Q_UNUSED(appdata_ptr);
	Q_UNUSED(numMsg);
	Q_UNUSED(msg);
	/*for (int i = 0; i < numMsg; i++)
	{
		qDebug("msg: %s", msg[i]->msg);
	}*/
	*resp = reply;

	return PAM_SUCCESS;
}
#endif

bool Session::authenticateSystemUser( QString username, QString password )
{
#ifdef Q_OS_WIN
	//LogonUser(L"LocalServer", L"NT AUTHORITY", NULL, LOGON32_LOGON_SERVICE, LOGON32_PROVIDER_DEFAULT, &hToken)
	return true;
#endif
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	reply = (struct pam_response *)malloc(sizeof(struct pam_response));

	const struct pam_conv local_conversation = { functionConversation, NULL };


	int retval;
	retval = pam_start("check_user", username.toUtf8().data(), &local_conversation, &local_auth_handle);

	if (retval != PAM_SUCCESS)
	{
		return false;
	}

	reply[0].resp = strdup( password.toUtf8().data() );
	reply[0].resp_retcode = 0;
	retval = pam_authenticate(local_auth_handle, 0);
	pam_set_item(local_auth_handle, PAM_RUSER, username.toUtf8().data());

	if (retval != PAM_SUCCESS)
	{
		return false;
	}

	return true;
#endif
}

QString Session::getUserHome( QString username )
{
	QString ret;
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	struct passwd pwd;
	struct passwd *result;
	char *buf;
	size_t bufsize;
	int s;

	bufsize = sysconf(_SC_GETPW_R_SIZE_MAX);
	if (bufsize == (size_t)-1)          /* Value was indeterminate */
	{
		bufsize = 16384;        /* Should be more than enough */
	}

	buf = (char*)malloc(bufsize);
	if (buf == NULL)
	{
		qDebug("Error getting memory for buffer");
		return QString::null;
	}

	s = getpwnam_r(username.toUtf8().data(), &pwd, buf, bufsize, &result);
	if (result == NULL)
	{
		if (s == 0)
		{
			qDebug("Not found\n");
		} else {
			qDebug("Error %d on getpwnam_r", s);
		}
	} else {
		ret = QString(result->pw_dir);
	}
	free( buf );
#endif
	return ret;
}

QString Session::createHome( QString homes, QString username )
{
	qDebug("Switching to user %s on %s...", username.toUtf8().data(), homes.toUtf8().data());
	QString homeBase = homes;
	if ( ! homeBase.endsWith(QDir::separator()) )
	{
		homeBase += QDir::separator();
	}
	// Get uid and gid fror username
	int uid, gid;
	getUserAndGroupIds(username, &uid, &gid);

	if ( uid == -1 || gid == -1 )
	{
		qDebug("Error obtaining user id/gid");
		return QString::null;
	}

	// Create home dir
	if ( ! QDir(homes).mkdir(username) )
	{
		qDebug("Warning: Cannot create home dir");
	}
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	if ( ! chown( (homeBase + username).toUtf8().data(), uid, gid) != 0 )
	{
		qDebug("Warning: Cannot change home dir owner");
	}
#endif
	return homeBase + username;
}

bool Session::switchUser( QString homeDir, QString username )
{
	// Get uid and gid fror username
	int uid, gid;
	getUserAndGroupIds(username, &uid, &gid);

	if ( uid == -1 || gid == -1 )
	{
		qDebug("Error obtaining user id/gid");
		return false;
	}

#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	// Create /run/user dir
	if ( ! QDir("/run/user/").mkdir( QString::number(uid) ) )
	{
		qDebug("Warning: Cannot create home dir");
	}
	if ( ! chown( ("/run/user/" + QString::number(uid)).toUtf8().data(), uid, gid) != 0 )
	{
		qDebug("Warning: Cannot change /run/user owner");
	}
	if ( ! chmod( ("/run/user/" + QString::number(uid)).toUtf8().data(), S_IRUSR | S_IWUSR | S_IXUSR) != 0 )
	{
		qDebug("Warning: Cannot change /run/user permissions");
	}
#endif

#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	// Get DISPLAY variable
	QByteArray display = qgetenv("DISPLAY");

	// Get number of user groups
	int ngroups = 1;
	gid_t * groups;
	groups = (gid_t*)malloc(ngroups * sizeof (gid_t));
	getgrouplist(username.toUtf8().data(), uid, groups, &ngroups);
	free(groups);

	// Get user group list
	groups = (gid_t*)malloc(ngroups * sizeof (gid_t));
	getgrouplist(username.toUtf8().data(), uid, groups, &ngroups);

	// Set principal group
	if ( setgid(gid) != 0 )
	{
		qDebug("Error: cannot set group id to %d", gid);
	}

	// Set group list
	setgroups(ngroups, groups);
	free(groups);

	// Set user id
	if ( setuid(uid) != 0 )
	{
		qDebug("Error: cannot set user id to %d", uid);
		return false;
	}

	// Open ConsoleKit session and create XDG_SESSION_COOKIE variable
	QByteArray ssid;
	QDBusInterface dbusManager( "org.freedesktop.ConsoleKit", "/org/freedesktop/ConsoleKit/Manager", "org.freedesktop.ConsoleKit.Manager", QDBusConnection::systemBus() );
	if ( dbusManager.isValid() )
	{
		QDBusMessage msg = dbusManager.call( "OpenSession" );
		if ( ! msg.errorName().isEmpty() )
		{
			qDebug("Cannot open session. Error: %s", msg.errorName().toUtf8().data());
		} else {
			QList<QVariant> args = msg.arguments();
			if ( args.count() >= 1 )
			{
				qputenv("XDG_SESSION_COOKIE", args[0].toByteArray() );
				msg = dbusManager.call( "GetSessionForCookie" );
				if ( ! msg.errorName().isEmpty() )
				{
					qDebug("Cannot get session for cookie %s. Error: %s", args[0].toByteArray().data(), msg.errorName().toUtf8().data());
				} else {
					QList<QVariant> sargs = msg.arguments();
					if ( sargs.count() >= 1 )
					{
						ssid = sargs[0].toByteArray();
					} else {
						qDebug("No response from system bus to get session for cookie.");
					}
				}
			} else {
				qDebug("No response from system bus to open session.");
			}
		}
	} else {
		qDebug("Cannot connect to system bus. Cannot open session.");
	}

	// Activate ConsoleKit session
	if ( ! ssid.isEmpty() )
	{
		QDBusInterface dbusSession( "org.freedesktop.ConsoleKit", "/org/freedesktop/ConsoleKit/Seat", "org.freedesktop.ConsoleKit.Seat", QDBusConnection::systemBus() );
		if ( dbusSession.isValid() )
		{
			QDBusMessage msg = dbusSession.call( "ActivateSession", ssid );
			if ( ! msg.errorName().isEmpty() )
			{
				qDebug("Cannot activate session. Error: %s", msg.errorName().toUtf8().data());
			}
		} else {
			qDebug("Cannot connect to system bus. Cannot activate session.");
		}
	}

	// Set environment variables
	QString shell;
	if ( QFile::exists("/bin/bash") )
	{
		shell = "/bin/bash";
	} else {
		shell = "/bin/sh";
	}

	QStringList paths;
	paths += "/usr/local/bin";
	paths += "/usr/bin";
	paths += "/bin";

	qputenv("SHELL", shell.toUtf8());
	qputenv("PATH", paths.join(":").toUtf8());
	qputenv("USER", username.toUtf8());
	qputenv("UID", QString::number( uid ).toUtf8());
	qputenv("HOME", homeDir.toUtf8());
	qputenv("XDG_RUNTIME_DIR", ("/run/user/" + QString::number( uid )).toUtf8());
	qputenv("DISPLAY", display);
	setCurrentDir( homeDir.toUtf8().data());

	QProcess * dbusLaunch = new QProcess();
	dbusLaunch->start("dbus-launch");
	if ( dbusLaunch->waitForStarted() && dbusLaunch->waitForFinished() )
	{
		while ( dbusLaunch->canReadLine() )
		{
			QString line = QString( dbusLaunch->readLine() );
			QString variable = line.section("=", 0, 0);
			QString content = line.right( line.size() - variable.size() - 1);
			qputenv( variable.toUtf8().data(), content.toUtf8() );
		}
	}

	qDebug("Finished user switching.");
	return true;
#endif
	return false;
}

void Session::syncGoogleChromeDir(QString username)
{
	QString baseDir = QStandardPaths::writableLocation( QStandardPaths::TempLocation ) + QDir::separator() + username;
	QString sourceDir = baseDir + QDir::separator() + ".config" + QDir::separator() + "google-chrome";
	QString targetDir = baseDir + QDir::separator() + ".librework.mnt" +  QDir::separator() + ".config" + QDir::separator() + "google-chrome";

	QStringList filesToSync;
	filesToSync += "Default/Preferences";
	filesToSync += "Default/Secure Preferences";
	filesToSync += "Default/History";
	filesToSync += "Default/History-journal";
	filesToSync += "Default/Shortcuts";
	filesToSync += "Default/Shortcuts-journal";
	filesToSync += "Default/Cookies";
	filesToSync += "Default/Cookies-journal";
	filesToSync += "Default/Login Data";
	filesToSync += "Default/Login Data-journal";
	filesToSync += "Default/Network Action Predictor";
	filesToSync += "Default/Network Action Predictor-journal";
	filesToSync += "Default/Web Data";
	filesToSync += "Default/Web Data-journal";

	for ( int i = 0; i < filesToSync.count(); i++ )
	{
		QString sourceFile = sourceDir + "/.config/google-chrome/" + filesToSync[i];
		QString targetFile = targetDir + "/.config/google-chrome/" + filesToSync[i];
		if ( QFile::exists( sourceFile ) )
		{
			QString command = "cp";
			QStringList arguments;
			arguments += "-a";
			arguments += sourceFile;
			arguments += targetFile;
			QProcess::execute( command, arguments );
		}
	}
}

void Session::createGoogleChromeDir( QString sourceDir )
{
	QString command = "tar";
	QStringList arguments;
	arguments += "-zxf";
	arguments += SHAREDIR "/google-chrome.tar.gz";
	arguments += "-C";
	arguments += sourceDir + "/.config";
	QProcess::execute( command, arguments );
}

void Session::createHomeLinks(QString sourceDir, QString homeDir, QString username)
{
	int uid, gid;
	getUserAndGroupIds(username, &uid, &gid);

#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	if ( ! QFile::exists( sourceDir + QDir::separator() + "google-chrome" ) )
	{
		createGoogleChromeDir( sourceDir );
	}
#endif

	QStringList syncDirs;
	syncDirs << ".config";

	QDir source( sourceDir );
	QStringList sourceList = source.entryList( QDir::NoDotAndDotDot );
	for ( int i = 0; i < sourceList.count(); i++)
	{
		if ( syncDirs.contains( sourceList[i] ) )
		{
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
			QString command = "cp";
			QStringList arguments;
			arguments += "-a";
			arguments += sourceDir + QDir::separator() + sourceList[i];
			arguments += homeDir + QDir::separator() + sourceList[i];
#endif
#ifdef Q_OS_WIN
			QString command = "xcopy";
			QStringList arguments;
			arguments += sourceDir + QDir::separator() + sourceList[i];
			arguments += homeDir + QDir::separator() + sourceList[i];
			arguments += "/S";
#endif
			QProcess::execute( command, arguments );
		} else {
			bool ret = QFile::link( sourceDir + QDir::separator() + sourceList[i], homeDir + QDir::separator() + sourceList[i] );
			if ( ret )
			{
				chown( QString(homeDir + QDir::separator() + sourceList[i] ).toUtf8().data(), uid, gid );
			}
		}
	}
}

void Session::createDesktopLinks(QString homeDir, QString username, bool setPermissions)
{
	int uid, gid;
	if ( setPermissions )
	{
		getUserAndGroupIds(username, &uid, &gid);
	}

	QStringList folders;
	folders << "Downloads" << "Documents" << "Desktop";
	QDir home( homeDir );
	for ( int i = 0; i < folders.count(); i++ )
	{
		if ( ! QFile::exists(homeDir + QDir::separator() + folders[i]) )
		{
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
			bool ret = home.mkdir( folders[i] );
			if ( ret && setPermissions )
			{
				chown( QString(homeDir + QDir::separator() + folders[i]).toUtf8().data(), uid, gid );
			}
#endif
			if ( folders[i] == "Desktop" )
			{
				QStringList links;
				links << "Downloads" << "Documents";
				for ( int i = 0; i < links.count(); i++ )
				{
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
					bool ret = QFile::link( homeDir + QDir::separator() + links[i], homeDir + QDir::separator() + folders[i] + QDir::separator() + links[i]);

					if ( ret && setPermissions )
					{
						chown( QString(homeDir + QDir::separator() + folders[i] + QDir::separator() + links[i]).toUtf8().data(), uid, gid );
					}
#endif
				}
			}
		}
	}
}

bool Session::openSession()
{
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	int retval = pam_open_session(local_auth_handle, PAM_SILENT);

	return (retval == PAM_SUCCESS);
#endif
	return true;
}

bool Session::closeSession()
{
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	int retval = pam_close_session(local_auth_handle, PAM_SILENT);
	retval = pam_end(local_auth_handle, retval);

	return (retval == PAM_SUCCESS);
#endif
	return true;
}

bool Session::connectToServer(QString serverAddress, int serverPort)
{
	if ( LWClient::isConnected() )
	{
		return true;
	}

	if ( LWClient::connect( serverAddress, serverPort ) )
	{
		return true;
	} else {
		return false;
	}
}

QStringList Session::launch(QString username, QString password, QString serverAddress, int port, LoginScreen * parent, QString remoteServer )
{
	QString homeDir = QString::null;
	QStringList error;
	QStringList warning;
	QString switchUser = username;
	QString token;

	Logger::logInfo( QString( "Launching user %1 session" ).arg( username ) );

	if ( ! remoteServer.isNull() && port <= 0 )
	{
		Logger::logInfo( "Remote session without Librework Server" );
		homeDir = LIBREWORK_HOME;
		switchUser= LIBREWORK_USER;

		Logger::logInfo( QString( "Mounting home for default librework user" ) );
		umountHome();
		mountHome();



		if ( Settings::getValue("EnableRemoteDesktopAuthCheck").toBool() )
		{
			if ( ! ClientSession::authenticateRemoteUser( username, password, serverAddress ) )
			{
				// Bad authentication
				error << tr("Bad authentication") << tr("Either the user name or the password you provided are invalid");
				Logger::logError( "Bad authentication: Either the user name or the password you provided are invalid" );
				return error;
			}
		}
	} else if ( port > 0 ) { // Auth against Librework server, for PXE or remote with Librework authentication
		Logger::logInfo( QString( "Session against server %1, port %2" ).arg( serverAddress ).arg( port ) );
		if ( connectToServer(serverAddress, port) )
		{
			Logger::logInfo( "Authenticating user..." );
			if ( LWClient::authenticate( username, password ) )
			{
				Logger::logInfo( "Authentication succesful. Connecting to Librework server." );
				token = LWClient::sendMessage( "get_user_token?" + username );
				bool ok;
				int uid = LWClient::sendMessage( "get_user_uid" ).toInt( &ok );
				if ( ok )
				{
					if ( ! Session::createSystemUser( uid, QStandardPaths::writableLocation( QStandardPaths::TempLocation ), username, password ) )
					{
						// Cannot create system user
						error << tr("Cannot create user") << tr("Cannot enable user with id %1 on this system").arg(uid);
						return error;
					}

					homeDir = Session::createHome( QStandardPaths::writableLocation( QStandardPaths::TempLocation ), username );
					if ( homeDir.isNull() )
					{
						error << tr("Cannot create home directory")<< tr("Cannot create home dir for user with id %1 on this system").arg(uid);
						return error;
					}

					if ( remoteServer.isNull() || Settings::getValue("RemoteDesktopMountRemoteShares").toBool()  )
					{
						if ( ! mountSmbHome( username, password ) )
						{
							warning << tr("Cannot mount home directory") << tr("WARNING: If you continue, this session data will not be saved on exit. Cannot mount home user folder, so a temporal directory will be created. Do you want to continue opening session?");
							if ( QMessageBox::warning( 0, warning[0], warning[1], QMessageBox::Yes | QMessageBox::No ) == QMessageBox::No )
							{
								error << tr("Cannot mount home directory") << tr("Cancelled user session");
								return error;
							}
						}
					}
					if ( Settings::getValue("RemoteDesktopMountRemoteShares").toBool() )
					{
						if ( ! mountSmbShares( username, password ) )
						{
							Logger::logInfo( "Cannot mount shared directory, you will not have access to shared folders on this session" );
						}
					}

					writeToken( username, homeDir, token );
				} else {
					// Cannot get system user uid
					error << tr("Cannot identify user") << tr("Cannot identify user %1 on this system").arg(username);
					return error;
				}
			} else {
				// Bad authentication
				error << tr("Bad authentication") << tr("Either the user name or the password you provided are invalid");
				Logger::logError( "Bad authentication: Either the user name or the password you provided are invalid" );
				return error;
			}
		} else {
			// Cannot connect to Librework Server
			error << tr("Bad connection") << tr("Cannot connect to Librework Server");
			Logger::logError( "Bad connection: Cannot connect to Librework Server" );
			return error;
		}
	} else { // No Librework Server, must authenticate against PAM or Samba
		QString profilesLocation = Settings::getValue("ProfilesLocation").toString();
		if ( ! profilesLocation.isEmpty() )
		{
			QString homes = QStandardPaths::writableLocation( QStandardPaths::TempLocation );
			Logger::logInfo( QString( "Creating system user %1" ).arg( username ) );
			if ( ! createSystemUser( -1, homes, username, password ) )
			{
				// Cannot create system user
				error << tr("Cannot create system user") << tr("Cannot create system user");
				Logger::logError( "Cannot create system user" );
				return error;
			}
			QString homeDir = createHome( homes, username );
			Logger::logInfo( QString( "Mounting home for user %1 at %2" ).arg( username ).arg( homeDir ) );
			if ( ! mountSmbHome( username, password ) )
			{
				// Bad authentication or cannot mount remote home
				error << tr("Bad authentication") << tr("Either the user name or the password you provided are invalid, or cannot access remote home dir");
				Logger::logError( "Bad authentication: Either the user name or the password you provided are invalid, or cannot access remote home dir" );
				return error;
			}
		} else {
			Logger::logInfo( "Session against PAM on localhost" );
			if ( ! Session::authenticateSystemUser(username, password ) )
			{
				// Bad authentication
				error << tr("Bad authentication") << tr("Either the user name or the password you provided are invalid");
				Logger::logError( "Bad authentication: Either the user name or the password you provided are invalid" );
				return error;
			}
		}
		Logger::logInfo( "Authentication succesful. Opening session..." );
		homeDir = Session::getUserHome(username);
		if ( homeDir.isNull() )
		{
			// Cannot get user home dir
			error << tr("Cannot identify user home") << tr("Cannot identify home of user %1 on this system").arg(username);
			return error;
		}
		bool ret = Session::openSession();
		if ( ret )
		{
			Logger::logInfo("Session succesfully opened");
		} else {
			Logger::logError("Cannot open session");
		}
	}

	if ( homeDir.isNull() )
	{
		// Cannot get user home dir
		error << tr("Cannot identify user home") << tr("Cannot identify home of user %1 on this system").arg(username);
		Logger::logError( QString( "Cannot identify user home: Cannot identify home of user %1 on this system").arg(username) );
		return error;
	}

	QString oldLastLogin = Settings::getValue("LastLogin").toString();
	Settings::setValue( "LastLogin", username );

	Logger::logInfo( QString( "Switching to user %1" ).arg( switchUser ) );

	if ( Session::switchUser( homeDir, switchUser ) )
	{

		Logger::logInfo( "User switched succesfully. Launching Librework Desktop" );
		QProcess * process = new QProcess();
		QStringList arguments;
		QTemporaryFile authFile;
		authFile.setAutoRemove( false );
		if ( authFile.open() )
		{
			authFile.write( username.toUtf8() + "\n" );
			authFile.write( password.toUtf8() + "\n" );

			if ( ! remoteServer.isNull() )
			{
				authFile.write( remoteServer.toUtf8() + "\n" );
			} else {
				authFile.write( serverAddress.toUtf8() + "\n" );
				authFile.write( QString::number(port).toUtf8() + "\n" );
			}
			authFile.close();
			arguments += "-a";
			arguments += authFile.fileName();
		} else {
			Logger::logWarning( QString( "Cannot create temporary file for user %1").arg(username) );
		}
		

		if ( ! remoteServer.isNull() )
		{
			Logger::logInfo( "Session is remote" );
			arguments += "-r";
		}

		process->start( SESSION_MANAGER, arguments );

		if ( ! process->waitForStarted() )
		{
			error << tr("Cannot start Librework Desktop") << tr("Cannot start Librework Desktop for user %1").arg(username);
			delete process;
			Logger::logError( QString( "Cannot start Librework Desktop: Cannot start Librework Desktop for user %1").arg(username) );
			return error;
		}

		Logger::logInfo( "Librework Desktop started." );

		connect( process, SIGNAL( finished(int) ), parent, SLOT( finishSession(int) ) );
	} else {
		Settings::setValue( "LastLogin", oldLastLogin );
		error << tr("Cannot start user session") << tr("Cannot start a session for user %1").arg(username);
		Logger::logError( QString( "Cannot start user session: Cannot start a session for user %1").arg(username) );
		return error;
	}

	return error; // No error
}

QString Session::credentialsFile(QString username, QString password)
{
	QDir tmp( QStandardPaths::writableLocation( QStandardPaths::TempLocation ) );
	tmp.mkdir( ".librework" );
	QString fileName = QStandardPaths::writableLocation( QStandardPaths::TempLocation )
				+ QString(QDir::separator())
				+ ".librework"
				+ QDir::separator()
				+ ".credentials";
	QFile file(fileName);
	file.setPermissions(QFileDevice::ReadOwner | QFileDevice::WriteOwner);
	file.open(QFile::WriteOnly);
	file.write(QString("username=" + username + "\n").toUtf8());
	file.write(QString("password=" + password + "\n").toUtf8());
	file.close();

	return fileName;
}

bool Session::mountSmbHome(QString username, QString password)
{
	bool ret = false;

	smbUmountAll();

	QString credentials = credentialsFile( username, password );

	// Mount home dir at .homepath
	QString smbHomePath;
	QString profilesLocation = Settings::getValue("ProfilesLocation").toString();
	if ( ! profilesLocation.isEmpty() )
	{
		smbHomePath = profilesLocation + "/" + username;
	} else {
		QString serverAddress = LWClient::sendMessage("get_profiles_server");
		Logger::logInfo( QString( "Mounting smb shares. Server address: '%1'" ).arg( serverAddress ) );
		if ( ! serverAddress.simplified().isEmpty() )
		{
			QString profilePath = LWClient::sendMessage("get_profiles_path");
			smbHomePath = "//" + serverAddress.section(":", 0, 0) + profilePath + "/" + username;
		} else {
			// Remove credentials file
			QFile credsFile(credentials);
			credsFile.remove();
			credsFile.close();
			return false;
		}
	}

	QString baseClientPath = QStandardPaths::writableLocation( QStandardPaths::TempLocation ) + QDir::separator();
	QString targetHomePath = baseClientPath + username;

	Logger::logInfo( QString( "Mounting %1 on %2" ).arg( smbHomePath ).arg( targetHomePath ) );
	if (QFile::exists( targetHomePath ) && ! smbHomePath.isEmpty())
	{
		if ( smbMount( username, credentials, smbHomePath, targetHomePath ) )
		{
			qputenv("HOME", targetHomePath.toUtf8());
			qputenv("USER", username.toUtf8());
			//createHomeLinks( targetHomePath, targetHomePath, username );
			Logger::logInfo( "Mount succesful" );
			ret = true;
		} else {
			Logger::logError( "Mount failed" );
		}
	}

	// Remove credentials file
	//QFile credsFile(credentials);
	//credsFile.remove();
	//credsFile.close();

	return ret;
}

bool Session::mountSmbShares(QString username, QString password)
{
	// Mount shares
	bool ret = true;
	Logger::logInfo("Mounting shares...");
	QString baseClientPath = QStandardPaths::writableLocation( QStandardPaths::TempLocation ) + QDir::separator();
	QString targetHomePath = baseClientPath + username;
	QString targetPath = targetHomePath + "/.shares/";
	QDir("/").mkpath( targetPath );
	QString credentials = credentialsFile( username, password );
	if ( QFile::exists( targetPath ) )
	{
		QStringList drives = LWClient::sendMessage("get_redirect_drives", QStringList( username )).split(",");
		Logger::logInfo( QString( "get_redirect_drives response: '%1'" ).arg( drives.join(',') ) );
		for ( int i = 0; i < drives.count(); i++ )
		{
			if ( ! drives[i].simplified().isEmpty() )
			{
				QString clientPath = drives[i].trimmed().section('=', 0, 0).remove("\\").remove(":").toUpper();
				QString serverPath = drives[i].trimmed().section('=', 1, 1);

				QString mountPoint = serverPath;
				mountPoint = mountPoint.replace("\\", "/");
				mountPoint = mountPoint.section('/', -1) + " (" + clientPath + ")";
				Logger::logInfo( QString( "Trying to mount %1 on %2" ).arg( serverPath ).arg( targetPath + mountPoint ) );

				if ( serverPath.startsWith("\\\\") ) // Shared folder
				{
					serverPath.replace("\\", "/");
					QDir("/").mkpath( targetPath + mountPoint );
					if ( QFile::exists( targetPath + mountPoint ) )
					{
						Logger::logInfo( QString( "Mounting %1 on %2" ).arg( serverPath ).arg( targetPath + mountPoint ) );
						ret = smbMount( username, credentials, serverPath, targetPath + mountPoint);
						if ( ! ret )
						{
							QDir("/").rmdir( targetPath + mountPoint );
						}
					} else {
						Logger::logWarning( QString( "Cannot create mount point %1" ).arg( targetPath + mountPoint ) );
						ret = false;
					}
				} else { // Client drive redirect
					// Do nothing
					Logger::logInfo( "Client drive redirect. Doing nothing." );
				}
			}
		}
	} else {
		Logger::logWarning( QString ( "Cannot create network shares dir %1. Not mounting shares" ).arg( targetPath ) );
		ret = false;
	}

	// Remove credentials file
	QFile credsFile(credentials);
	credsFile.remove();
	credsFile.close();

	return ret;
}

bool Session::smbMount(QString username, QString credentialsFile, QString smbPath, QString targetPath)
{
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	int uid, gid;
	getUserAndGroupIds(username, &uid, &gid);

	QProcess smbMountProcess;
	QString command = "mount.cifs";
	QStringList parameters;
	QString options = Settings::getValue("Advanced/CIFSMountOptions").toString();

	if ( options.isEmpty() )
	{
		options = "mfsymlinks,nobrl,cache=strict,actimeo=60,nounix";
	}

	parameters << "-o";
	parameters << options + ",uid=" + QString::number(uid) + "," + "credentials=" + credentialsFile;
	parameters << smbPath;
	parameters << targetPath;

	Logger::logInfo( QString( "Mounting samba dir: %1 %2" ).arg( command ).arg( parameters.join(" ") ) );

	smbMountProcess.start( command, parameters );
	if ( smbMountProcess.waitForStarted() )
	{
		smbMountProcess.waitForFinished();
		return ( smbMountProcess.exitCode() == 0 );
	}
#endif
	return false;
}

void Session::smbUmountAll()
{
	QProcess * umount = new QProcess();
	QStringList parameters;
	parameters << "-t" << "cifs" << "-a";
	umount->start( "umount", parameters );
	umount->waitForStarted();
	umount->waitForFinished();
	delete umount;
}

void Session::mountHome()
{
	QProcess * mount = new QProcess();
	QStringList parameters;
	parameters << "-t" << "tmpfs" << "none" << LIBREWORK_HOME;
	mount->start( "mount", parameters );
	mount->waitForStarted();
	mount->waitForFinished();
	delete mount;
}

void Session::umountHome()
{
	QProcess * umount = new QProcess();
	QStringList parameters;
	parameters << LIBREWORK_HOME;
	umount->start( "umount", parameters );
	umount->waitForStarted();
	umount->waitForFinished();
	delete umount;
}

void Session::writeToken( QString username, QString homeDir, QString token )
{
	QString userConfigFile = homeDir + QDir::separator() + QString(USER_CONFIG_FILE_NAME);

#ifdef Q_OS_UNIX
	QStringList arguments;
	arguments += username;
	arguments += userConfigFile;
	QProcess::execute( "chown", arguments );
	qDebug("Command: 'chown %s'", arguments.join(" ").toUtf8().data());

	arguments.clear();
	arguments += "0600";
	arguments += userConfigFile;
	QProcess::execute( "chmod", arguments );
#endif

	QSettings desktopSettings( userConfigFile, QSettings::IniFormat );
	desktopSettings.setValue( "UserToken", token );
}
