#include "sessionmanager.h"

Login1Manager SessionManager::login1Manager;

int SessionManager::setCurrentDir(const char *dirname)
{
	QDir::setCurrent(dirname);
#ifdef Q_OS_WIN
	if (SetCurrentDirectoryA(dirname))
	{
		return 0;
	} else {
		return -1;
	}
#endif

#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	return chdir(dirname);
#endif
}

bool SessionManager::switchUser( QString homeDir, QString username )
{
	// Get uid and gid fror username
	int uid, gid;
	SysUser::getUserAndGroupIds(username, &uid, &gid);

	if ( uid == -1 || gid == -1 )
	{
		qDebug("Error obtaining user id/gid");
		return false;
	}

#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	// Create /run/user dir
	QDir("/run/").mkdir( "user" );
	if ( ! QDir("/run/user/").mkdir( QString::number(uid) ) )
	{
		Logger::logWarning(tr("Cannot create /run/user/%1 dir").arg(uid));
	}
	if ( chown( ("/run/user/" + QString::number(uid)).toUtf8().data(), uid, gid) != 0 )
	{
		Logger::logWarning(tr("Cannot change /run/user/%1 owner").arg(uid));
	}
	if ( chmod( ("/run/user/" + QString::number(uid)).toUtf8().data(), S_IRUSR | S_IWUSR | S_IXUSR) != 0 )
	{
		Logger::logWarning(tr("Cannot change /run/user/%1 permissions").arg(uid));
	}
	// Get DISPLAY variable
	QByteArray display = qgetenv("DISPLAY");

	// Get number of user groups
	int ngroups = 1;
	gid_t * groups;
	groups = (gid_t*)malloc(ngroups * sizeof (gid_t));
	getgrouplist(username.toUtf8().data(), uid, groups, &ngroups);
	free(groups);

	// Get user group list
	groups = (gid_t*)malloc(ngroups * sizeof (gid_t));
	getgrouplist(username.toUtf8().data(), uid, groups, &ngroups);

	// Set principal group
	if ( setgid(gid) != 0 )
	{
		Logger::logWarning(tr("Cannot set group id to %1").arg(gid));
	}

	// Set group list
	setgroups(ngroups, groups);
	free(groups);

	// Set user id
	if ( setuid(uid) != 0 )
	{
		Logger::logError(tr("Cannot set user id to %1").arg(uid));
		return false;
	}

	// Open ConsoleKit session and create XDG_SESSION_COOKIE variable
	QByteArray ssid;
	QDBusInterface dbusManager( "org.freedesktop.ConsoleKit", "/org/freedesktop/ConsoleKit/Manager", "org.freedesktop.ConsoleKit.Manager", QDBusConnection::systemBus() );
	if ( dbusManager.isValid() )
	{
		QDBusMessage msg = dbusManager.call( "OpenSession" );
		if ( ! msg.errorName().isEmpty() )
		{
			Logger::logWarning(tr("Cannot open session. Error: %1").arg( msg.errorName()));
		} else {
			QList<QVariant> args = msg.arguments();
			if ( args.count() >= 1 )
			{
				qputenv("XDG_SESSION_COOKIE", args[0].toByteArray() );
				msg = dbusManager.call( "GetSessionForCookie" );
				if ( ! msg.errorName().isEmpty() )
				{
					Logger::logWarning(tr("Cannot get session for cookie %1. Error: %2").arg(args[0].toByteArray().data()).arg( msg.errorName()));
				} else {
					QList<QVariant> sargs = msg.arguments();
					if ( sargs.count() >= 1 )
					{
						ssid = sargs[0].toByteArray();
					} else {
						Logger::logWarning("No response from system bus to get session for cookie.");
					}
				}
			} else {
				Logger::logWarning("No response from system bus to open session.");
			}
		}
	} else {
		Logger::logWarning("Cannot connect to system bus. Cannot open session.");
	}

	// Activate ConsoleKit session
	if ( ! ssid.isEmpty() )
	{
		QDBusInterface dbusSession( "org.freedesktop.ConsoleKit", "/org/freedesktop/ConsoleKit/Seat", "org.freedesktop.ConsoleKit.Seat", QDBusConnection::systemBus() );
		if ( dbusSession.isValid() )
		{
			QDBusMessage msg = dbusSession.call( "ActivateSession", ssid );
			if ( ! msg.errorName().isEmpty() )
			{
				Logger::logWarning(tr("Cannot activate session. Error: %1").arg(msg.errorName()));
			}
		} else {
			Logger::logWarning("Cannot connect to system bus. Cannot activate session.");
		}
	}

	// Set environment variables
	QString shell;
	if ( QFile::exists("/bin/bash") )
	{
		shell = "/bin/bash";
	} else {
		shell = "/bin/sh";
	}

	QStringList paths;
	paths += "/usr/local/bin";
	paths += "/usr/bin";
	paths += "/bin";

	qputenv("SHELL", shell.toUtf8());
	qputenv("PATH", paths.join(":").toUtf8());
	qputenv("USER", username.toUtf8());
	qputenv("UID", QString::number( uid ).toUtf8());
	qputenv("HOME", homeDir.toUtf8());
	qputenv("XDG_RUNTIME_DIR", ("/run/user/" + QString::number( uid )).toUtf8());
	qputenv("DISPLAY", display);
	setCurrentDir( homeDir.toUtf8().data());

	QProcess * dbusLaunch = new QProcess();
	dbusLaunch->start("dbus-launch");
	if ( dbusLaunch->waitForStarted() && dbusLaunch->waitForFinished() )
	{
		while ( dbusLaunch->canReadLine() )
		{
			QString line = QString( dbusLaunch->readLine() );
			QString variable = line.section("=", 0, 0);
			QString content = line.right( line.size() - variable.size() - 1);
			qputenv( variable.toUtf8().data(), content.toUtf8() );
		}
	}

	Logger::logInfo("Finished user switching.");
	return true;
#endif
	return false;
}

bool SessionManager::connectToServer(QString serverAddress, int serverPort)
{
	if ( LWClient::isConnected() )
	{
		return true;
	}

	if ( LWClient::connect( serverAddress, serverPort ) )
	{
		return true;
	} else {
		return false;
	}
}

bool SessionManager::authenticateSessionUser( QString username, QString password, SessionSetting sessionSetting )
{
	Logger::logInfo( QString( "Authenticating user %1 with authentication type %2" ).arg( username ).arg( sessionSetting.authentication().toLower() ) );
	if ( sessionSetting.authentication().toLower() == "none" )
	{
		return true;
	}

	if ( sessionSetting.authentication().toLower() == "local" )
	{
		bool ret = Pam::authenticateSystemUser( username, password );
		if ( ret )
		{
			Logger::logDebug("Authentication successful");
		} else {
			Logger::logDebug("Something failed authenticating");
		}
		return ret;
	}

	if ( sessionSetting.authentication().contains( "://" ) )
	{
		QString protocol = sessionSetting.authentication().section("://", 0, 0);
		QString server = sessionSetting.authentication().section("://", 1, 1);
		Logger::logInfo( QString( "Authentication Server: %1" ).arg( sessionSetting.authentication() ) );
		QString domain;
		if ( server.contains('@') )
		{
			domain = server.section('@', 0, 0);
			server = server.section('@', 1, 1);
		}
		if ( protocol.toLower() == "smb" ) // Active Directory or Samba authentication
		{
			return Shares::smbAuth( username, password, domain, server );
		}

		if ( protocol.toLower() == "rdp" ) // RDP authentication
		{
			QString fullUsername = username;
			if ( ! domain.isEmpty() )
			{
				fullUsername = username + "@" + domain;
			}
			Logger::logInfo( QString( "Authenticating user with RDP username: %1, server: %2" ).arg( fullUsername ).arg( server ) );
			return ClientSession::authenticateRemoteUser( fullUsername, password, server );
		}

		if ( protocol.toLower() == "lw" ) // Librework Server authentication
		{
			QString serverAddress = server;
			int serverPort = 4445;
			if ( server.contains( ":" ) )
			{
				serverAddress = server.section( ":", 0, 0 );
				serverPort = server.section( ":", 1, 1 ).toInt();
			}
				
			if ( connectToServer(serverAddress, serverPort) )
			{
				Logger::logInfo( "Authenticating user..." );
				return LWClient::authenticate( username, password );
			}
		}
	}

	return false;
}

bool SessionManager::checkWhiteList(QString username, SessionSetting sessionSetting)
{
	return sessionSetting.whiteList().contains( username ) || sessionSetting.whiteList().count() == 0;
}

bool SessionManager::checkBlackList(QString username, SessionSetting sessionSetting)
{
	return sessionSetting.blackList().contains( username );
}

QStringList SessionManager::launch(QString username, QString password, SessionSetting sessionSetting, LoginScreen * parent )
{
	QString homeDir = QString::null;
	QStringList error;
	QString token;

	if ( ! checkWhiteList( username, sessionSetting ) )
	{
		// Not in white list
		QString errorTitle = tr("White list checking");
		QString errorText = tr("This user ('%1') has no permissions to launch this session").arg(username);
		error << errorTitle << errorText;
		Logger::logError( QString( errorTitle + ":" + errorText ) );
		return error;
	}

	if ( checkBlackList( username, sessionSetting ) )
	{
		// In black list
		QString errorTitle = tr("Black list checking");
		QString errorText = tr("This user ('%1') has no permissions to launch this session").arg(username);
		error << errorTitle << errorText;
		Logger::logError( QString( errorTitle + ":" + errorText ) );
		return error;
	}

	bool ret = true;

	if ( ! authenticateSessionUser( username, password, sessionSetting ) )
	{
		Logger::logInfo( QString("User %1 log in failed").arg(username) );
		// Bad authentication
		QString errorTitle = tr("Bad authentication");
		QString errorText = tr("Either the user name or the password you provided are invalid");
		error << errorTitle << errorText;
		Logger::logError( QString( errorTitle + ":" + errorText ) );
		ret = false;
	} else {
		Logger::logInfo( QString("User %1 succesfully logged in").arg(username) );
	}

	if ( Pam::closeSession( true ) ) // Close greeter session
	{
		Logger::logInfo("Greeter session succesfully closed");
	} else {
		Logger::logError("Cannot close greeter session");
	}

	if ( ret )
	{
		//ret = Pam::stopPam( true, 0 );
		if ( ! ret )
		{
			Logger::logWarning("Something happened stopping PAM for greeter session");
		}
		ret = Pam::startPam( false, "librework-displaymanager", username );
	}

	if ( ! ret )
	{
		Logger::logError("Cannot start pam user session");
	} else {
		ret = Pam::openSession( "user", "seat0" ); // Open user session
	}

	if ( ret )
	{
		Logger::logInfo("User session succesfully opened");

		QString session;
		QList<Session> sessionList = login1Manager.listSessions();
		for ( int i = 0; i < sessionList.count(); i++ )
		{
			Logger::logDebug( QString( "Session name: %1, Session user: '%2', username: '%3'" ).arg( sessionList[i].session ).arg( sessionList[i].username ).arg( username ) );
			if ( sessionList[i].username == username )
			{
				session = sessionList[i].session;
			}
		}
		if ( ! session.isEmpty() )
		{
			if ( ! login1Manager.activateSession( session ) )
			{
				// Cannot activate session
				QString errorTitle = tr("Bad user activation");
				QString errorText = tr("Cannot activate session on seat 0 for user %1").arg( username );
				error << errorTitle << errorText;
				Logger::logError( QString( errorTitle + ":" + errorText ) );
				ret = true;
			}
		} else {
			// Bad user session
			QString errorTitle = tr("Bad user session");
			QString errorText = tr("There is no session for this user");
			error << errorTitle << errorText;
			Logger::logError( QString( errorTitle + ":" + errorText ) );
			ret = true;
		}
	} else {
		Logger::logError("Cannot open user session");
	}

	if ( ! ret )
	{
		bool r = Pam::closeSession( false ); // Close user session if opened
		r = Pam::stopPam( false, 0 );
		/*r = Pam::startPam( true, "librework-displaymanager-greeter", "root" );
		if ( ! r )
		{
			Logger::logError("Cannot start pam root session for greeter");
		} else {
			r = Pam::openSession( "greeter", "seat0" ); // Open user session again
		}*/
		return error;
	}

	QString homePath;
	int uid = -1;
	int gid = -1;
	if ( sessionSetting.isLocal() )
	{
		homeDir = SysUser::getUserHome( username );
		SysUser::getUserAndGroupIds( username, &uid, &gid );
	} else {
		homePath = QStandardPaths::writableLocation( QStandardPaths::TempLocation );
	}

	if ( sessionSetting.isLibrework() )
	{
		token = LWClient::sendMessage( "get_user_token?" + username );
		bool ok;
		uid = LWClient::sendMessage( "get_user_uid" ).toInt( &ok );
		if ( ! ok )
		{
			uid = -1;
		}
	}
	if ( ! sessionSetting.isLocal() && ! SysUser::createSystemUser( uid, homePath, username, password ) )
	{
		QString errorTitle = tr("Cannot create session user");
		QString errorText = tr("Cannot create session user %1 on this system").arg(username);
		error << errorTitle << errorText;
		Logger::logError( QString( errorTitle + ":" + errorText ) );
		return error;
	}

	if ( ! sessionSetting.isLocal() )
	{
		homeDir = homePath + QDir::separator() + username;
		if ( ! QFileInfo( homeDir ).isDir() )
		{
			if ( ! QDir( homePath ).mkdir( username ) )
			{
				homeDir = QString(); // Cannot create home dir
			}
		}
	}

	if ( homeDir.isEmpty() )
	{
		QString errorTitle = tr("Cannot create home directory");
		QString errorText = tr("Cannot create home dir for user %1 on this system").arg(username);
		error << errorTitle << errorText;
		Logger::logError( QString( errorTitle + ":" + errorText ) );
		return error;
	}

	QList<ShareSetting> shareSettings = sessionSetting.shares();
	Logger::logInfo(QString("Getting shares settings for %1: %2 entries").arg( sessionSetting.name() ).arg( shareSettings.count() ) );
	// First, mount home dir
	for ( int i = 0; i < shareSettings.count(); i++ )
	{
		if ( shareSettings[i].isHome() )
		{
			Logger::logInfo( QString("Home from %1%2").arg( shareSettings[i].serverAddress() ).arg( shareSettings[i].sourcePath() ) );
			if ( ! Shares::mountShare(username, password, shareSettings[i].domain(), homeDir, shareSettings[i] ) )
			{
				QString wTitle = tr("Cannot mount home directory");
				QString wText = tr("WARNING: If you continue, this session data will not be saved on exit. Cannot mount home user folder, so a temporal directory will be created. Do you want to continue opening session?");
				Logger::logInfo( QString( wTitle + ":" + wText ) );
				if ( QMessageBox::warning( 0, wTitle, wText, QMessageBox::Yes | QMessageBox::No ) == QMessageBox::No )
				{
					QString errorTitle = tr("Cannot mount home directory");
					QString errorText = tr("Cancelled user session");
					error << errorTitle << errorText;
					Logger::logError( QString( errorTitle + ":" + errorText ) );
					return error;
				}
			}
		}
	}
	if ( shareSettings.count() > 0 )
	{
		// Empty targetDir
		QDir targetDir( shareSettings[0].targetDir() );
		QFileInfoList list = targetDir.entryInfoList();
		for (int i = 0; i < list.count(); i++)
		{
			QFileInfo fileInfo = list.at(i);
			if ( fileInfo.isDir() )
			{
				QString target = fileInfo.absoluteFilePath();
				Logger::logInfo( QString("Looking on %1 for remove").arg( target ) );
				if ( QDir( target ).isEmpty() )
				{
                                        QDir( target ).removeRecursively();
                                } else {
					Logger::logInfo( QString("Not empty") );
				}
			} else {
				Logger::logInfo( QString("Not dir %1").arg(fileInfo.absoluteFilePath()) );
			}
		}
	}

	// Then mount shares
	for ( int i = 0; i < shareSettings.count(); i++ )
	{
		if ( ! shareSettings[i].isHome() )
		{
			Logger::logInfo( QString("Share from %1%2").arg( shareSettings[i].serverAddress() ).arg( shareSettings[i].sourcePath() ) );

			if ( ! Shares::mountShare( username, password, shareSettings[i].domain(), homeDir, shareSettings[i] ) )
			{
				QString wTitle = tr("Cannot mount user share");
				QString wText = tr("Cannot mount user share %1 for user %2").arg(shareSettings[i].sourcePath()).arg(username);
				Logger::logInfo( QString( wTitle + ":" + wText ) );
				/*if ( QMessageBox::warning( 0, wTitle, wText, QMessageBox::Yes | QMessageBox::No ) == QMessageBox::No )
				{
					QString errorTitle = tr("Cannot mount user share");
					QString errorText = tr("Cancelled user session");
					error << errorTitle << errorText;
					Logger::logError( QString( errorTitle + " - " + errorText ) );
					return error;
				}*/
			}
		}
	}


	QString oldLastLogin = Settings::getValue("LastLogin").toString();
	Settings::setValue( "LastLogin", username );
	if ( ! switchUser( homeDir, username ) )
	{
		Settings::setValue( "LastLogin", oldLastLogin );
		QString errorTitle = tr("Cannot switch to session user");
		QString errorText = tr("Cannot switch to session user %1").arg(username);
		error << errorTitle << errorText;
		Logger::logError( QString( errorTitle + ":" + errorText ) );
		return error;
	}

	if ( ! launchSessionDesktop(username, password, sessionSetting, parent) )
	{
		Settings::setValue( "LastLogin", oldLastLogin );
		QString errorTitle = tr("Cannot start user session");
		QString errorText = tr("Cannot start a session for user %1").arg(username);
		error << errorTitle << errorText;
		Logger::logError( QString( errorTitle + ":" + errorText ) );
		return error;
	}

	Login1Manager login1Manager;

	QString sessionsMessage;
        QList<Session> sessions = login1Manager.listSessions();
        for ( int i = 0; i < sessions.count(); i++ )
        {
                sessionsMessage += sessions[i].session + " " + sessions[i].username + " " + sessions[i].seat + "\n";
        }

        QString seatsMessage;
        QList<Seat> seats = login1Manager.listSeats();
        for ( int i = 0; i < seats.count(); i++ )
        {
                seatsMessage += seats[i].seat + " " + seats[i].path.path() + "\n";
        }

	//QMessageBox::information( nullptr, tr("Debug"), "Sessions:\n" + sessionsMessage + "\nSeats:\n" + seatsMessage);

	return QStringList(); // No error
}

bool SessionManager::launchSessionDesktop(QString username, QString password, SessionSetting sessionSetting, LoginScreen * parent)
{
	QStringList error;

	Logger::logInfo( "User switched succesfully. Launching Librework Desktop" );
	QProcess * process = new QProcess();
	QStringList arguments;
	QTemporaryFile authFile;
	authFile.setAutoRemove( false );
	if ( authFile.open() )
	{
		authFile.write( username.toUtf8() + "\n" );
		authFile.write( password.toUtf8() + "\n" );

		if ( ! sessionSetting.isRemote() )
		{
			authFile.write( sessionSetting.name().toUtf8() + "\n" );
		} else {
			if ( sessionSetting.clientSession().isEmpty() )
			{
				QString serverAddress = sessionSetting.desktop().toLower();
				int port = 3389;
				if ( serverAddress.contains( "rdp://" ) )
				{
					serverAddress = serverAddress.section( "rdp://", 1, 1 );
				}
				if ( serverAddress.contains( ":" ) )
				{
					port = serverAddress.section( ":", 1, 1 ).toInt();
					serverAddress = serverAddress.section( ":", 0, 0 );
				}
				authFile.write( serverAddress.toUtf8() + "\n" );
				authFile.write( QString::number(port).toUtf8() + "\n" );
			} else {
				authFile.write( sessionSetting.clientSession().toUtf8() + "\n" );
			}
		}
		authFile.close();
		arguments += "-a";
		arguments += authFile.fileName();
	} else {
		Logger::logWarning( QString( "Cannot create temporary file for user %1").arg(username) );
	}

	if ( sessionSetting.isRemote() )
	{
		Logger::logInfo( "Session is remote" );
		arguments += "-r";
	}

	connect( process, SIGNAL( finished(int) ), parent, SLOT( finishSession(int) ) );

	process->start( SESSION_MANAGER, arguments );

	if ( ! process->waitForStarted() )
	{
		error << tr("Cannot start Librework Desktop") << tr("Cannot start Librework Desktop for user %1").arg(username);
		delete process;
		Logger::logError( QString( "Cannot start Librework Desktop: Cannot start Librework Desktop for user %1").arg(username) );
		return false;
	}

	Logger::logInfo( "Librework Desktop started." );

	return true;
}

void SessionManager::mountHome()
{
	QProcess * mount = new QProcess();
	QStringList parameters;
	parameters << "-t" << "tmpfs" << "none" << LIBREWORK_HOME;
	mount->start( "mount", parameters );
	mount->waitForStarted();
	mount->waitForFinished();
	delete mount;
}

void SessionManager::umountHome()
{
	QProcess * umount = new QProcess();
	QStringList parameters;
	parameters << LIBREWORK_HOME;
	umount->start( "umount", parameters );
	umount->waitForStarted();
	umount->waitForFinished();
	delete umount;
}
