#include "sysuser.h"

void SysUser::getUserAndGroupIds( QString username, int * uid, int * gid )
{
#ifdef Q_OS_WIN
	Q_UNUSED( username );
	Q_UNUSED( uid );
	Q_UNUSED( gid );
#endif
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	struct passwd pwd;
	struct passwd *result;
	char *buf;
	size_t bufsize;
	int s;

	bufsize = sysconf(_SC_GETPW_R_SIZE_MAX);
	if (bufsize == (size_t)-1)          /* Value was indeterminate */
	{
		bufsize = 16384;        /* Should be more than enough */
	}

	buf = (char*)malloc(bufsize);
	if (buf == NULL)
	{
		* uid = -1;
		* gid = -1;
		qDebug("Error getting memory for buffer");
		return;
	}

	s = getpwnam_r(username.toUtf8().data(), &pwd, buf, bufsize, &result);
	if (result == NULL)
	{
		if (s == 0)
		{
			qDebug("Not found\n");
		} else {
			qDebug("Error %d on getpwnam_r", s);
		}
		* uid = -1;
		* gid = -1;
	} else {
		* uid = result->pw_uid;
		* gid = result->pw_gid;
	}
	free( buf );
#endif
} 

bool SysUser::createSystemUser( int uid, QString homes, QString username, QString password )
{
#ifdef Q_OS_WIN
	Q_UNUSED( uid );
	Q_UNUSED( homes );
	Q_UNUSED( username );
	Q_UNUSED( password );
#endif
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	int _uid = uid;
	bool ret = false;
	bool exists = false;
	int lastUserId = 1000;
	QFile passwd(LIBREWORK_PASSWD_FILE);
	if (passwd.open(QIODevice::ReadOnly))
	{
		QTextStream in(&passwd);
		while (!in.atEnd())
		{
			QString line = in.readLine();
			QString lineUsername = line.section(":", 0, 0);
			int lineUserId = line.section(":", 2, 2).toInt();
			if ( lineUserId > lastUserId )
			{
				lastUserId = lineUserId;
			}
			if ( lineUserId == _uid )
			{
				exists = true;
				if ( lineUsername == username )
				{
					ret = true;
				}
			}
		}
		passwd.close();
	}

	if ( _uid == -1 )
	{
		_uid = lastUserId + 1;
	}

	if ( ! exists )
	{
		QFile passwd(LIBREWORK_PASSWD_FILE);
		passwd.open(QIODevice::WriteOnly | QIODevice::Append);
		QByteArray line;
		line += username;
		line += ":x:"; // No password entry
		line += QString::number( _uid );
		line += ":50::"; // Staff group
		line += homes + "/" + username;
		line += ":/bin/bash\n";
		if ( passwd.write( line ) == -1 )
		{
			Logger::logError( QString( "Cannot write to file %1" ).arg( LIBREWORK_PASSWD_FILE ) );
			ret = false;
		} else {
			ret = true;
		}
		passwd.close();

		QStringList libreworkUserGroups = Settings::getValue("Advanced/LibreworkNSSUserGroups").toStringList();
		QFile lwGroup(LIBREWORK_GROUP_FILE);
		if ( ret && lwGroup.open(QIODevice::WriteOnly | QIODevice::Append) )
		{
			QFile group("/etc/group");
			if ( group.open(QIODevice::ReadOnly))
			{
				QTextStream in(&group);
				while ( ret && ! in.atEnd() )
				{
					QString line = in.readLine();
					QString lineGroupName = line.section(":", 0, 0);
					QString lineGroupId = line.section(":", 2, 2);
					if ( libreworkUserGroups.contains( lineGroupName ) )
					{
						QByteArray groupLine = QString( lineGroupName + ":x:" + lineGroupId + ":" + username + "\n" ).toUtf8();
						if ( lwGroup.write( groupLine ) == -1 )
						{
							Logger::logError( QString( "Cannot write to file %1" ).arg( LIBREWORK_GROUP_FILE ) );
							ret = false;
						}
					}
				}
				group.close();
			}
			lwGroup.close();
		}
		QFile lwShadow(LIBREWORK_SHADOW_FILE);
		if ( ret && lwShadow.open(QIODevice::WriteOnly | QIODevice::Append) )
		{
			lwShadow.setPermissions( QFileDevice::ReadOwner );
			QByteArray line = username.toUtf8() + ":" + QCryptographicHash::hash( password.toUtf8(), QCryptographicHash::Sha512 ) + ":19999:0:99999:7:::\n";
			if ( lwShadow.write( line ) == -1 )
			{
				Logger::logError( QString( "Cannot write to file %1" ).arg( LIBREWORK_SHADOW_FILE ) );
				ret = false;
			} else {
				ret = true;
			}
			lwShadow.close();
		}
	}

	return ret;
#endif
	return false;
}

QString SysUser::getUserHome( QString username )
{
	QString ret;
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	struct passwd pwd;
	struct passwd *result;
	char *buf;
	size_t bufsize;
	int s;

	bufsize = sysconf(_SC_GETPW_R_SIZE_MAX);
	if (bufsize == (size_t)-1)          /* Value was indeterminate */
	{
		bufsize = 16384;        /* Should be more than enough */
	}

	buf = (char*)malloc(bufsize);
	if (buf == NULL)
	{
		qDebug("Error getting memory for buffer");
		return QString::null;
	}

	s = getpwnam_r(username.toUtf8().data(), &pwd, buf, bufsize, &result);
	if (result == NULL)
	{
		if (s == 0)
		{
			qDebug("Not found\n");
		} else {
			qDebug("Error %d on getpwnam_r", s);
		}
	} else {
		ret = QString(result->pw_dir);
	}
	free( buf );
#endif
	return ret;
}

QString SysUser::createHome( QString homes, QString username )
{
	qDebug("Switching to user %s on %s...", username.toUtf8().data(), homes.toUtf8().data());
	QString homeBase = homes;
	if ( ! homeBase.endsWith(QDir::separator()) )
	{
		homeBase += QDir::separator();
	}
	// Get uid and gid fror username
	int uid, gid;
	getUserAndGroupIds(username, &uid, &gid);

	if ( uid == -1 || gid == -1 )
	{
		qDebug("Error obtaining user id/gid");
		return QString::null;
	}

	// Create home dir
	if ( ! QDir(homes).mkdir(username) )
	{
		qDebug("Warning: Cannot create home dir");
	}
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
	if ( ! chown( (homeBase + username).toUtf8().data(), uid, gid) != 0 )
	{
		qDebug("Warning: Cannot change home dir owner");
	}
#endif
	return homeBase + username;
}

void SysUser::createHomeLinks(QString sourceDir, QString homeDir, QString username)
{
	int uid, gid;
	getUserAndGroupIds(username, &uid, &gid);

	QStringList syncDirs;
	syncDirs << ".config";

	QDir source( sourceDir );
	QStringList sourceList = source.entryList( QDir::NoDotAndDotDot );
	for ( int i = 0; i < sourceList.count(); i++)
	{
		if ( syncDirs.contains( sourceList[i] ) )
		{
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
			QString command = "cp";
			QStringList arguments;
			arguments += "-a";
			arguments += sourceDir + QDir::separator() + sourceList[i];
			arguments += homeDir + QDir::separator() + sourceList[i];
#endif
#ifdef Q_OS_WIN
			QString command = "xcopy";
			QStringList arguments;
			arguments += sourceDir + QDir::separator() + sourceList[i];
			arguments += homeDir + QDir::separator() + sourceList[i];
			arguments += "/S";
#endif
			QProcess::execute( command, arguments );
		} else {
			bool ret = QFile::link( sourceDir + QDir::separator() + sourceList[i], homeDir + QDir::separator() + sourceList[i] );
			if ( ret )
			{
				chown( QString(homeDir + QDir::separator() + sourceList[i] ).toUtf8().data(), uid, gid );
			}
		}
	}
}

void SysUser::createDesktopLinks(QString homeDir, QString username, bool setPermissions)
{
	int uid, gid;
	if ( setPermissions )
	{
		getUserAndGroupIds(username, &uid, &gid);
	}

	QStringList folders;
	folders << "Downloads" << "Documents" << "Desktop";
	QDir home( homeDir );
	for ( int i = 0; i < folders.count(); i++ )
	{
		if ( ! QFile::exists(homeDir + QDir::separator() + folders[i]) )
		{
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
			bool ret = home.mkdir( folders[i] );
			if ( ret && setPermissions )
			{
				chown( QString(homeDir + QDir::separator() + folders[i]).toUtf8().data(), uid, gid );
			}
#endif
			if ( folders[i] == "Desktop" )
			{
				QStringList links;
				links << "Downloads" << "Documents";
				for ( int i = 0; i < links.count(); i++ )
				{
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD)
					bool ret = QFile::link( homeDir + QDir::separator() + links[i], homeDir + QDir::separator() + folders[i] + QDir::separator() + links[i]);

					if ( ret && setPermissions )
					{
						chown( QString(homeDir + QDir::separator() + folders[i] + QDir::separator() + links[i]).toUtf8().data(), uid, gid );
					}
#endif
				}
			}
		}
	}
}

void SysUser::writeToken( QString username, QString homeDir, QString token )
{
	QString userConfigFile = homeDir + QDir::separator() + QString(USER_CONFIG_FILE_NAME);

#ifdef Q_OS_UNIX
	QStringList arguments;
	arguments += username;
	arguments += userConfigFile;
	QProcess::execute( "chown", arguments );
	qDebug("Command: 'chown %s'", arguments.join(" ").toUtf8().data());

	arguments.clear();
	arguments += "0600";
	arguments += userConfigFile;
	QProcess::execute( "chmod", arguments );
#endif

	QSettings desktopSettings( userConfigFile, QSettings::IniFormat );
	desktopSettings.setValue( "UserToken", token );
}
